#!/usr/bin/python3
import numpy


class BaseWriter:
  def __init__(self, filename):
    self.f = open(filename,"wb")

class MultiSpectrumWriter(BaseWriter):
  def __init__(self, filename):
    super().__init__(filename)
    self.data = []
    self.names = []
    self.units = []
    self.identifiers = []

  def addX(self, data, identifier):
    self.data.append(data.values)
    self.names.append(data.datatype.name)
    self.units.append(data.datatype.unit)
    self.identifiers.append(identifier)
    if data.errors is not None:
      self.data.append(data.errors)
      self.names.append(data.datatype.name)
      self.units.append(data.datatype.unit)
      self.identifiers.append(identifier+"_Error")

  def addY(self, data, identifier):
    if len(data.values) < len(self.data[0]):
      vals = numpy.zeros(self.data[0].shape)
      vals[0:len(data.values)] = data.values
    else:
      vals = data.values
    self.data.append(vals)   
    self.names.append(data.datatype.name)
    self.units.append(data.datatype.unit)
    self.identifiers.append(identifier)
    if data.errors is not None:
      self.data.append(data.errors)
      self.names.append(data.datatype.name)
      self.units.append(data.datatype.unit)
      self.identifiers.append(identifier+"_Error")
 
  def addMultiY(self, Ys, labels, commonlabel=""):
    for i,Y in enumerate(Ys):
      self.data.append(Y)
      self.identifiers.append("{0}".format(labels[i])+commonlabel)

  def addSpectrum(self, spectrum, identifier="Y"):
    self.addX(spectrum.axis, "X")
    self.addY(spectrum.data, identifier)

  def addSpectra(self, spectra):
    self.addX(spectra.axis, "X")
    self.addMultiY(spectra.data.values, spectra.keys.values)
  
  def addFit(self, spectrum, fit, convolve=False):
    if convolve:
      y = spectrum.evaluateConvolutionFit(fit)
    else:
      y = fit.evaluate(self.data[0])
    self.addY(y, fit.__class__.__name__)

  def addFits(self, fit):
    self.addMultiY(fit.batchEvaluate(self.data[0]), self.headers[1:], fit.__class__.__name__)

  def write(self, fmt="%4f"):
    data = numpy.array(self.data).T
    self.f.write(numpy.compat.asbytes("\t".join(self.identifiers)+"\n"))
    self.f.write(numpy.compat.asbytes("\t".join(self.names)+"\n"))
    self.f.write(numpy.compat.asbytes("\t".join(self.units)+"\n"))
    numpy.savetxt(self.f, data, fmt=fmt, delimiter="\t")
    self.f.close()

class SpectrumWithFitWriter(MultiSpectrumWriter):
  def __init__(self, filename, spectrum, fit, withIRF=False):
    super().__init__(filename)
    self.addSpectrum(spectrum, identifier="Data")
    if withIRF: 
      self.addY(spectrum.IRF, identifier="IRF")
    self.addFit(spectrum, fit, withIRF)
    self.write()
   




#!/usr/bin/python3
import numpy
import pickle
from fepydas.datatypes.Data import Data1D,Data2D,DataType
import fepydas.datatypes as datatypes
from fepydas.datatypes.Dataset import SpectraWithCommonIRF,Spectrum,SpectrumWithIRF,SpectrumSeries,PLE
from fepydas.constructors.datareaders.Proprietary import BeckerHicklHistogram, PicoHarpHistogram, TimeHarpHistogram


def GenericXY(x,xname,xunit,y,yname,yunit):
  return Spectrum(Data1D(x,DataType(xname,xunit)),Data1D(x,DataType(xname,xunit)))


def GenericXMultiY(x,xname,xunit,y,yname,yunit,keys,keyname,keyunit):
  return SpectrumSeries(Data1D(x,DataType(xname,xunit)),Data1D(keys,DataType(keyname,keyunit)),Data2D(y,DataType(yname,yunit)))


def ASCII(filename, delimiter="\t",missing_values='', skip_header=0):
  data = numpy.genfromtxt(filename,delimiter=delimiter,missing_values=missing_values,skip_header=skip_header)
  print("ASCII File {0}. Shape:{1}".format(filename,data.shape))
  return data

def ASCII_SpectrumRowSeries(filename, xType:DataType, keyType:DataType, valueType:DataType):
  data = ASCII(filename)
  xAxis = Data1D(data[0,:],xType)
  keyAxis = Data1D(numpy.arange(len(data[:,0])-1),keyType)
  vals = Data2D(data[1:,:],valueType)
  return SpectrumSeries(xAxis,keyAxis,vals)

def ASCII_Histogram(filename,resolution):
  data = ASCII(filename,delimiter=" ",skip_header=10)
  xAxis = Data1D(numpy.arange(len(data))*resolution,datatypes.TIME_ns)
  yAxis = Data1D(data,datatypes.COUNTS)
  return Spectrum(xAxis,yAxis)

def Binary(filename):
    return pickle.load(open(filename,"rb"))

def HistogramReader(filename):
    print("Reading {0}".format(filename))
    filetype = filename[-3:]
    if filetype=="phu":
       time, data = PicoHarpHistogram(filename)
    elif filetype=="thi":
       time, data = TimeHarpHistogram(filename)
    elif filetype=="sdt":
       time, data = BeckerHicklHistogram(filename)
    else:
       print("Unsupported Histogram Type: .{0}".format(filetype))
    axis = Data1D(time,datatypes.TIME_ns)
    data = Data1D(data,datatypes.COUNTS)
    return axis, data

def Histogram(filename):
    axis, data = HistogramReader(filename)
    return Spectrum(axis,data)    

def HistogramWithIRF(filename,filenameIRF):
    histoTime, histoData = HistogramReader(filename)
    IRFTime, IRFData = HistogramReader(filenameIRF)
    if numpy.sum(histoTime.values-IRFTime.values)>0:
        print("Warning, IRF and Decay Histogram Time Axes not compatible!!")
    return SpectrumWithIRF(histoTime, histoData, IRFData)

def HistogramsWithCommonIRF(filenames, filenameIRF):
    IRFTime, IRFData = HistogramReader(filenameIRF)
    data = numpy.ndarray(shape=(len(filenames),len(IRFTime.values)))
    for i, n in enumerate(filenames):
      histoTime, histoData = HistogramReader(n)
      if numpy.sum(histoTime.values-IRFTime.values)>0:
        print("Warning, IRF and Decay Histogram Time Axes not compatible for {0}!!".format(n))
      data[i,:] = histoData.values
      filenames[i] = filenames[i].split("/")[-1]
    CombinedData = Data2D(data,IRFData.datatype)
    return SpectraWithCommonIRF(IRFTime, CombinedData, IRFData, Data1D(filenames,DataType("Filename","")))



#!/usr/bin/python3
from copy import deepcopy
import numpy
import pickle

from fepydas.datatypes.Data import Data, Data1D, Data2D, Data3D, Response, VacuumCorrection
from fepydas.datatypes import NUMBER
from fepydas.workers.Fit import LimitedGaussianFit, GaussianFit, CalibrationFit, Fit, LinearFit, AutomaticCalibration
from fepydas.workers.ASCIIWriter import MultiSpectrumWriter

from sklearn.decomposition import NMF, PCA, LatentDirichletAllocation
from sklearn.cluster import SpectralClustering

from scipy.interpolate import griddata

class BaseDataset:
  def saveBinary(self, filename):
    f = open(filename,"bw")
    pickle.dump(self,f)
    f.close()

class Dataset(BaseDataset):
  def __init__(self, axis: Data1D, data: Data):
    self.axis = axis
    self.data = data

  def applyAxisTransformation(self, transformation):
    self.axis.values = transformation.apply(self.axis.values)
    print("Applied Axis Transformation")
    
  def applyVacuumCorrection(self, temperature=20, pressure=101325, humidity=0, co2=610):
    vac = VacuumCorrection(temperature, pressure, humidity, co2)
    self.axis.values = vac.apply(self.axis.values)

  def convertToEnergy(self, unit="eV"):
    self.axis.convertToEnergy(unit)
    self.data.applyJacobianTransformation(self.axis)

  def antiBloom(self, over, width=1, reverse=False):
    self.data.antiBloom(over, width=width, reverse=reverse)

  def cutAxis(self, minVal, maxVal):
    idx1, idx2 = self.axis.cutAndGetIndexRange(minVal,maxVal)
    self.data.cut(idx1,idx2)

  def cutAxisByIdx(self, minIdx, maxIdx):
    self.axis.cut(minIdx,maxIdx)
    self.data.cut(minIdx,maxIdx)

  def divideBySpectrum(self, spectrum):
    self.data.divideBy(spectrum.data.values)

class Spectrum(Dataset):
  def __init__(self, axis: Data1D, data: Data1D):
    super().__init__(axis,data)

  def export(self,filename):
    msw = MultiSpectrumWriter(filename)
    msw.addSpectrum(self)
    msw.write()

  def identifyPeaks(self, threshold=10, width=10):
    derivative = numpy.diff(self.data.values)
    derivative.resize(self.data.values.shape)
    nonflat = derivative != 0
    zerocrossings = numpy.diff(numpy.sign(derivative)) != 0
    zerocrossings.resize(self.data.values.shape)
    noiseLevel = self.data.values > numpy.average(self.data.values)*threshold
    peaks = zerocrossings * noiseLevel * nonflat
    print("{0} peaks detected".format(numpy.sum(peaks)))
    indeces = numpy.where(peaks)[0]
    idxRanges = []
    for i, idx in enumerate(indeces):
      if idx+width>len(self.axis.values) or idx-width<0:
        continue
      idxRanges.append([idx-width, idx+width])
    return idxRanges
    
  def normalize(self):
    self.data.normalize()

  def toPlot(self):
    from fepydas.constructors.Plots import SpectrumPlot
    return SpectrumPlot(self)

  def toScatterPlot(self):
    from fepydas.constructors.Plots import SpectrumScatterPlot
    return SpectrumScatterPlot(self)

class SpectrumSeries(Dataset):
  def __init__(self, axis:Data1D, keys:Data1D, data:Data2D):
    super().__init__(axis,data)
    self.keys = keys
  
  def average(self):
    return Spectrum(self.axis, self.data.average())

  def collapse(self, filter=None):
    return Spectrum(self.axis, self.data.collapse(filter))

  def collapseDominatingCluster(self, maxClusters = None):
    averages, cluster, resp = self.performGMM(maxClusters=maxClusters)
    dominatingCluster = numpy.bincount(cluster).argmax()
    keep = numpy.where(cluster==dominatingCluster)
    return self.collapse(filter=keep)

  def cutKeys(self, min, max):
    keys = numpy.where((self.keys.values>=min) & (self.keys.values<=max))[0]
    self.keys.values = self.keys.values[keys]
    self.data.values = self.data.values[keys,:]
                       
  def export(self, filename):
    writer = MultiSpectrumWriter(filename)
    writer.addSpectra(self)
    writer.write()

  def filterByIntegral(self, threshold=0.5):
    integrals = self.data.integrate()
    max = numpy.amax(integrals.values)
    filter = numpy.where(integrals.values/max > threshold)[0]
    self.data.values = self.data.values[filter,:]
    self.keys.values = self.keys.values[filter]
    
  def findDominatingComponents(self, n = 1, algo="NMF"):
    if algo=="NMF":
      decomp = NMF(n_components = n)
    elif algo=="PCA":
      decomp = PCA(n_components = n)
    elif algo=="LDA":
      decomp = LatentDirichletAllocation(n_components = n)
    return self.fitDominatingComponents(decomp)
    
  def fit(self, fitter:Fit):
    fitter.batchFit(self.axis.values, self.data.values)
    
  def fitDominatingComponents(self, decomp):
    decomp.fit(self.data.values)
    return decomp

  def getDataValues(self):
    return self.data.values
  
  def fitMixture(self, n=2, nmf=None):
    data = self.getDataValues()
    

    if nmf is not None:
      #pcaWorker = NMF()
      #pcaWorker.fit(data)
      #np = numpy.where(pcaWorker.explained_variance_ratio_ > pca)[0][-1]
      data = NMF(n_components = nmf).fit_transform(data)
      #print("PCA reduced complexity to {0} components".format(len(pcaWorker.explained_variance_ratio_)))

    aff = numpy.zeros(shape=(data.shape[0],data.shape[0]))
    for i in range(data.shape[0]):
      for j in range(data.shape[0]):
        aff[i,j] = numpy.dot(data[i,:],data[j,:])

    gmm = SpectralClustering(affinity="precomputed", n_jobs=-1, eigen_solver="amg", n_clusters=n)
    gmm.fit(aff)

    return gmm

  def getDominatingComponent(self, algo="PCA"):
    decomp = self.findDominatingComponents(n=1, algo=algo)
    return Spectrum(self.axis, Data1D(decomp.components_[0], self.data.datatype))

  def getDominatingComponents(self, algo="PCA", n=2):
    decomp = self.findDominatingComponents(n=n, algo=algo)
    return decomp, SpectrumSeries(self.axis, Data1D(range(n), NUMBER), Data2D(decomp.components_, self.data.datatype))

  def getProjection(self, decomp):
    return SpectrumSeries(self.keys, Data1D(range(len(decomp.components_)), NUMBER), Data2D(decomp.transform(self.data.values).T, self.data.datatype))

  def removeComponents(self, decomp):
    self.data.values -= decomp.transform(self.data.values) @ decomp.components_
  
  def removeDominatingComponents(self, n=1, algo="NMF"):
    decomp = self.findDominatingComponents(n=n, algo=algo)
    self.removeComponents(decomp)
    return decomp

  def highDynamicRange(self, max=0.99, zmin=None):
    if self.data.values.shape[0] == 1:
      #nothing to do here
      return Spectrum(deepcopy(self.axis), Data1D(self.data.values.flatten(), self.data.datatype))
    #calculate weights
    zmax = (numpy.amax(self.data.values)*max) 
    if zmin is None:
      zmin = (numpy.amin(self.data.values))
    zfloor = numpy.full_like(self.data.values, zmin)
    zceil = numpy.full_like(self.data.values, zmax)
    zcenter, zrange = (zmax+zmin)/2, (zmax-zmin)/2
    weights =(zrange - numpy.abs((numpy.maximum(numpy.minimum(self.data.values,zceil),zfloor) - zcenter))) #Low weight for low (under-exposed) or high (over-exposed) values, high weight for medium values

    #calculate mappings
    num = len(self.keys.values)
    crossweights = numpy.zeros((num,num,len(self.data.values[0,:])))
    for i in range(num):
      for j in range(num):
        crossweights[i,j,:] = numpy.multiply(weights[i,:],weights[j,:])
    integratedCWs = numpy.sum(crossweights, axis=2)
    idxs = numpy.flip(numpy.argsort(numpy.sum(integratedCWs, axis=1)))
  
    
    def linearProjection(dataIn, dataRef, crossweights):
      return dataIn*numpy.average(dataRef/dataIn, weights=crossweights)

    mapped = numpy.zeros((num))
    mapped[idxs[0]] = 1

    for i in range(len(idxs)-1):
      idx = idxs[i+1]
      usableCWs = numpy.multiply(integratedCWs[idx,:],mapped)
      best = numpy.argsort(usableCWs)[-1]
      self.data.values[idx,:] = linearProjection(self.data.values[idx,:],self.data.values[best,:], crossweights[idx,best,:])
      print("mapped",idx,"to",best)
      mapped[idx]=1

    return Spectrum(deepcopy(self.axis), Data1D(numpy.average(self.data.values, axis=0, weights=weights),self.data.datatype))

  def integrate(self):
    return Spectrum(self.keys, self.data.integrate())

  def interpolate(self, newKeys):
    oldPoints = numpy.ndarray(shape=(len(self.keys.values)*len(self.axis.values),2))
    values =numpy.ndarray(shape=(len(self.keys.values)*len(self.axis.values)))
    newPoints = numpy.ndarray(shape=(len(newKeys)*len(self.axis.values),2))

    n=0
    for it,t in enumerate(self.keys.values):
      for ix,x in enumerate(self.axis.values):
        oldPoints[n,:]=[t,x]
        values[n] = self.data.values[it,ix]
        n+=1
    n=0
    for it,t in enumerate(newKeys):
      for ix,x in enumerate(self.axis.values):
        newPoints[n,:]=[t,x]
        n+=1
    self.data.values = griddata(oldPoints, values, newPoints, method="cubic").reshape(len(newKeys),len(self.axis.values))
    self.keys.values = newKeys

  def maximum(self):
    return Spectrum(self.keys, Data1D(self.axis.values[numpy.argmax(self.data.values,axis=1)],self.axis.datatype))

  
  def normalize(self, individual=False):
    if individual:
      self.data.normalizeIndividual()
    else:
      self.data.normalize()

  def subtractBaseline(self):
    #Invoke only if no dark spectrum available!
    baseline = numpy.min(self.data.values)-1
    self.data.values-=(baseline)
    print("Baseline removed: {0}".format(baseline))

  def toPlot(self, polar=False):
    from fepydas.constructors.Plots import SpectrumSeriesPlot
    return SpectrumSeriesPlot(self, polar=polar)

class BinnedSpectrumSeries(SpectrumSeries):
  def __init__(self, spectrumSeries, binner, num_bins):
    keys = binner.createBins(spectrumSeries.keys, num_bins)
    data, self.histogram = binner.binData(spectrumSeries.data)
    super().__init__(deepcopy(spectrumSeries.axis),keys,data)

class JoinSpectrumSeries(SpectrumSeries):
  def __init__(self, spectra, interpol=False, names=None):
    if len(spectra)==1:
      if names: keys = [names[0]]
      else: keys = [0]
      data = numpy.ndarray(shape=(1, len(spectra[0].data.values)),dtype=spectra[0].data.values.dtype)
      data[0,:] = spectra[0].data.values
      super().__init__(spectra[0].axis, Data1D(keys,None), Data2D(data,spectra[0].data.datatype))
      return
    if names:
      keys = Data1D(names, NUMBER)
    else:
      keys = Data1D(range(len(spectra)),NUMBER)
    if interpol:
      valmin, mins, maxs, res = [], [],[],[]
      for i in range(len(spectra)):
        mins.append(numpy.amin(spectra[i].axis.values))
        maxs.append(numpy.amax(spectra[i].axis.values))
        res.append(numpy.average(numpy.diff(spectra[i].axis.values)))
        if res[i] < 0: #axis must be ascending
          spectra[i].axis.values = numpy.flip(spectra[i].axis.values)
          spectra[i].data.values = numpy.flip(spectra[i].data.values)
        valmin.append(numpy.amin(spectra[i].data.values))
      resolution = numpy.abs(numpy.average(res))
      min, max = numpy.amin(mins), numpy.amax(maxs)
      valmin = numpy.amin(valmin)
      #Interpolation, set non-overlapping regions to minimum so weight will be 0
      newAxis = Data1D(numpy.linspace(min, max, int((max-min)/resolution)), spectra[0].axis.datatype)
      data = Data2D(numpy.zeros((len(spectra),len(newAxis.values))), spectra[0].data.datatype)
      for i in range(len(spectra)):
        data.values[i,:] = numpy.interp(newAxis.values, spectra[i].axis.values, spectra[i].data.values, left=valmin, right=valmin) - valmin +0.001
      super().__init__(newAxis,keys,data)
    else:
      data = Data2D(numpy.zeros((len(spectra),len(spectra[0].data.values))), spectra[0].data.datatype)
      for i in range(len(spectra)):
        data.values[i,:] = spectra[i].data.values
      #TODO check AXIS consistency
      super().__init__(spectra[0].axis,keys,data)
    #TODO default create keys in SS init

class MergeSpectraSeries(SpectrumSeries):
  def __init__(self,spectraSeries):
    first = True
    for spectrumSeries in spectraSeries:
      if first:
        super().__init__(deepcopy(spectrumSeries.axis),deepcopy(spectrumSeries.keys),deepcopy(spectrumSeries.data))
        first = False
      else:
        self.axis.append(spectrumSeries.axis.values)
        self.data.append(spectrumSeries.data.values)
        self.keys.append(spectrumSeries.keys.values)


class Map(BaseDataset):
  def __init__(self, mapping:Data3D, data:Data2D):
    self.mapping = mapping
    self.data = data

  def export(self, filename):
    writer = MultiSpectrumWriter(filename)
    x,y = self.mapping.extractAxes()
    writer.addX(x,"")
    for i in range(len(y.values)):
      writer.addY(self.data.getCrosssectionAt(i),"{0}".format(y.values[i]))
    writer.write()

  def insertFlatData(self, data):
    x, y = self.data.values.shape
    print(x,y,data.shape)
    self.data.values = data.reshape(x,y)
    
  def toPlot(self, log=False, zRange=None):
    from fepydas.constructors.Plots import ContourMapPlot
    return ContourMapPlot(self, log=log, zRange=zRange)

    

class SpectrumSeriesWithCalibration(SpectrumSeries):
  def __init__(self, axis: Data1D, data:Data2D, calibration: Data1D):
    super().__init__(axis=axis,keys=None,data=data)
    self.calibrated = False
    self.calibration = calibration

  def getCalibrationSpectrum(self):
    return Spectrum(self.axis, self.calibration)

  def calibrate(self, references, threshold=10, width=10):
    self.applyAxisTransformation(AutomaticCalibration(self.getCalibrationSpectrum(), references, threshold=threshold, width=width).toTransformation())
    self.calibrated = True
    print("Calibrated")

class BaseSpectrumMap(SpectrumSeriesWithCalibration):
  def __init__(self, axis: Data1D, data:Data2D, mapping: Data3D, calibration: Data1D):
    super().__init__(axis, data, calibration)
    self.mapping = mapping
    
  def average(self):
    return Spectrum(self.axis, self.data.average())
  
  def getEmptyMap(self, datatype = NUMBER):
    x,y,z = self.mapping.values.shape
    vals = numpy.ndarray(shape=(x,y))
    values = Data2D(vals, datatype)
    return Map(self.mapping, values)

class SparseSpectrumMap(BaseSpectrumMap):
  def __init__(self, axis: Data1D, data: Data2D, mapping:Data3D, mask: Data2D, calibration:Data1D):
    super().__init__(axis,data,mapping,calibration)
    self.mask = mask #This contains the indexes of the data array where the invidiual spectra are stored or numpy.nan for empty pixels

  def cutMap(self, xMin, xMax, yMin, yMax):
    i1,i2,i3,i4 = self.mapping.cutAndGetIndexRange(xMin,xMax,yMin,yMax)
    self.mask.cutMap(i1,i2,i3,i4)
    self.updateFromMask()
    #the spectra are still there, but they are not referenced anymore. this is okay for most plots, but should be fixed if fits are attempted?
  
  def fitParameterToMap(self, fitter, paramName, paramDatatype):
    map = self.getEmptyMap(paramDatatype)
    x, y = map.data.values.shape
    for i in range(x):
      for j in range(y):
        idx = self.mask.values[i,j]
        result = fitter.results[idx]
        if result==0:
          val = 0
        else:
          val = result.params[paramName].value
        map.data.values[i,j] = val
    return map
    
  def integrate(self):
    return Map(self.mapping, self.data.integrate().to2D(self.mask))
    
  def maximum(self):
    return Map(self.mapping, Data1D(self.axis.values[numpy.argmax(self.data.values,axis=1)],self.axis.datatype).to2D(self.mask))
    
  def updateFromMask(self):
    #This method removes spectra that are not part of the map anymore and updates the mask 
    #TODO I probably need to check for NaN here?
    table = numpy.array(self.mask.values.flatten(),dtype=numpy.int)
    self.data.values = self.data.values[table,:]
    coordinates = [] #I need to temp store this so I don't accidently "find" a value post-replacement    
    for i,old in enumerate(table):
      coordinates.append(numpy.where(self.mask.values==old))      
    for i,pos in enumerate(coordinates):
      self.mask.values[pos] = i

class SpectrumMap(BaseSpectrumMap):
  def __init__(self, axis: Data1D, data:Data3D, mapping:Data3D, calibration: Data1D):
    super().__init__(axis,data,mapping,calibration)

  def cutMap(self, xMin, xMax, yMin, yMax):
    i1,i2,i3,i4 = self.mapping.cutAndGetIndexRange(xMin,xMax,yMin,yMax)
    self.data.cutMap(i1,i2,i3,i4)

  def getClusterMapIdx(self, gmm):
    map = self.getEmptyMap()
    #map.insertFlatData(gmm.predict(self.getDataValues(pca)))
    map.insertFlatData(gmm.labels_)
    return map

  def getClusterIntegratedSpectra(self, gmm):
    map = self.getClusterMapIdx(gmm)
    num = numpy.amax(map.data.values)+1
    data = numpy.zeros(shape=(num,self.data.values.shape[2]))
    for i in range(num):
      cond = numpy.where(map.data.values == i)
      filtered = self.data.values[cond]
      data[i,:] = numpy.sum(filtered, axis=0)
    return SpectrumSeries(self.axis, Data1D(range(num),NUMBER) ,Data2D(data, self.data.datatype))
    
  def export(self, filename):
    writer = MultiSpectrumWriter(filename)
    writer.addX(self.axis, "Coordinates")
    for x in range(self.mapping.values.shape[0]):
      for y in range(self.mapping.values.shape[0]):
        writer.addY(self.data.get1DAt(x,y), "{0} : {1}".format(self.mapping.values[x,y,0],self.mapping.values[x,y,1]))
    writer.write()

  def fit(self, fitter:Fit, filter=None):
    fitter.batchFit(self.axis.values, self.data.flatten(filter=filter))

  def fitDominatingComponents(self, decomp):
    decomp.fit(self.data.flatten())
    return decomp

  def fitParameterToMap(self, fitter, paramName, paramDatatype, filter=None):
    map = self.getEmptyMap(paramDatatype)
    x, y, z = self.data.values.shape
    if filter is not None:
      wasFitted = numpy.zeros_like(self.mapping.values[:,:,0])
      print(wasFitted.shape)
      wasFitted[filter] = 1
      print(wasFitted.shape)
      wasFitted = wasFitted.reshape((x*y,1))
      print(wasFitted.shape)
    vals = numpy.ndarray(shape=(x*y))
    i = 0
    for key in fitter.results.keys():
      if filter is not None:
        while wasFitted[i]==0:
          vals[i] = 0
          i+=1
      result = fitter.results[key]
      if result==0:
        vals[i] = 0
      else:
        vals[i] = result.params[paramName].value
      i+=1
    map.insertFlatData(vals)
    return map
  
  def getDataValues(self, pca=None):
    if pca is not None:
      return pca.transform(self.data.flatten())
    return self.data.flatten()

  def integrate(self):
    return Map(self.mapping, self.data.integrate())

  def maximum(self):
    return Map(self.mapping, Data2D(self.axis.values[numpy.argmax(self.data.values,axis=2)],self.axis.datatype))




class PLE(SpectrumSeriesWithCalibration):
  def __init__(self, axis: Data1D, keys:Data1D, data:Data2D, calibration = None, response = None):
    super().__init__(axis,data,calibration)
    self.keys = keys
    self.response = response
  
  def applyLampResponseCorrection(self,zeroFloor=False):
    interpolatedResponse = self.response.interpolateResponse(self.keys.values,zeroFloor)
    for i in range(len(self.keys.values)):
      self.data.values[i] /= interpolatedResponse[i]
    print("Lamp Response Correction made")

  
  def calibrateLamp(self,width=2,thresh=1):
    #This assumes the lamp is present in some of the spectra and used to calibrate the entire excitation (keys) axis with a linear fit
    fit = LimitedGaussianFit()
    nominal = []
    fitted = []
    fittedErrs = []
    for i in range(len(self.keys.values)):
      x, y = fit.initializeAutoLimited(self.axis.values,self.data.values[i],self.keys.values[i],width,thresh)
      if type(x) is numpy.ndarray and len(x)>40:
        #print(x,y)
        fit.fit(x,y)
        nominal.append(self.keys.values[i])
        fitted.append(fit.result.params["x0"].value)
        fittedErrs.append(fit.result.params["x0"].stderr)
    nominal = numpy.array(nominal,dtype=numpy.float64)
    fitted = numpy.array(fitted,dtype=numpy.float64)
    calib = CalibrationFit()
    calib.initializeAuto()
    calib.fit(nominal,fitted)#,peakErrs)
    print(calib.result.params)
    transformation = calib.toTransformation()
    self.keys.values = transformation.apply(self.keys.values)
    print("Calibrated Lamp")
    return transformation

  def calibrateLampExplicit(self, width=2, thresh=1):
    #This assumes the lamp is present in ALL spectra and is used to set the excitation (key) to the fitted value for each spectrum
    #The resulting axis may be used to calibrate another PLE measurement which used the same lamp parameters
    #USE WITH CAUTION AND CHECK RESULTS
    fit = GaussianFit()
    for i in range(len(self.keys.values)):
      idxs = numpy.where(numpy.abs(self.axis.values-self.keys.values[i])<=width)[0]
      x, y = self.axis.values[idxs], self.data.values[i,idxs]
      fit.initializeAuto(x,y)
      fit.fit(x,y)
      self.keys.values[i] = fit.result.params["x0"].value
      #print(lamp.keys.values[i], fit.parameters["x0"])
    print("Calibrated Lamp")
  
  def export(self, filename):
    writer = MultiSpectrumWriter(filename)
    writer.addSpectra(self)
    writer.write()
    

  def toPlot(self, log=True):
    from fepydas.constructors.Plots import ContourPlot
    return ContourPlot(self, log=log)

class SpectrumWithIRF(Spectrum):
  def __init__(self, axis: Data1D, data: Data, IRF: Data1D):
    super().__init__(axis,data)
    self.trimIRF(IRF)

  def evaluateConvolutionFit(self, fit):
    return Data1D(fit.evaluateConvolution(self.axis.values, self.IRF.values), self.data.datatype)

  def getExtendedIRF(self):
    data = numpy.zeros(self.data.values.shape)
    for i,x in enumerate(self.IRF.values):
      data[self.IRFStart+i]=x
    return data/data.max()

  def trimIRF(self,IRF):
    self.IRFStart = 0#lastZero+1
    self.IRF = IRF
    self.IRF.trim()
     
  def deconvolve(self,eps=None):
    self.data.deconvolve(self.IRF,eps)

class SpectraWithCommonIRF(SpectrumWithIRF):
  def __init__(self, axis: Data1D, data: Data2D, IRF: Data1D, identifiers: Data1D):
    super().__init__(axis,data,IRF)
    self.keys = identifiers

  def normalize(self, individual=False):
    if individual:
      self.data.normalizeIndividual()
    else:
      self.data.normalize()



import numpy
import os
from fepydas.datatypes.Dataset import Spectrum, PLE, SpectrumSeries, SparseSpectrumMap
from fepydas.constructors.datareaders.Generic import ASCII
from fepydas.datatypes.Data import Data1D, Data2D, Data3D, DataType
import fepydas.datatypes as datatypes

def Automatic(filename):
  file = open(filename, encoding="Latin-1")
  done = False
  headers = {}
  while not done:
    line = file.readline().strip().split(" = ")
    #print(line)
    if line[0]=="HEADER":
      done = True
    else:
      headers[line[0]] = line[1:]
  #Header Done
  if headers["Measurement Type"][0] == "Mapscan Cube":
    #For Mapscans, the Header is followed by a section showing the mapping using X and O.
    map_x = int(headers["Map Dimension X"][0])
    map_y = int(headers["Map Dimension Y"][0])
    map_x0 = float(headers["Map Edge Position X (µm)"][0])
    map_y0 = float(headers["Map Edge Position Y (µm)"][0])
    map_pitch = float(headers["Map Pitch (µm)"][0])
    file.readline() #---
    mapping = numpy.ndarray(shape=(map_x,map_y,2), dtype=numpy.float)
    bool_mask = numpy.ndarray(shape=(map_x,map_y), dtype=bool)
    mask = numpy.ndarray(shape=(map_x,map_y), dtype=numpy.float)
    num = 0
    for i in range(map_y):
      line = file.readline()
      for j in range(map_x):
        mapping[j,map_y-1-i]=[map_x0+j*map_pitch, map_y0+(map_y-1-i)*map_pitch]
        if (line[j]=="X"):
          bool_mask[j,map_y-1-i] = True
          num+=1
        else:
          bool_mask[j,map_y-1-i] = False
    file.readline() #---
    
    names = file.readline().strip().split("\t")
    units = file.readline().strip().split("\t")
    
    endOfAxis = False
    axisData = []
    while not endOfAxis:
      value = file.readline().strip()
      if value == "---":
        break
      axisData.append(float(value.split("\t")[0]))
    axis = Data1D(numpy.array(axisData),DataType(names[0],units[0]))
    
    data = numpy.ndarray(shape=(num, len(axis.values)))
    raw = numpy.fromfile(file, dtype = numpy.dtype(">u8"))[1:] #64bit uint big-endian
    #It seems that this array is transposed?
    data = raw.reshape((len(axis.values), num)).T
    n = 0
    for i in range(map_x):
      for j in range(map_y):
        if bool_mask[i,j]:
          mask[i,j] = n
          #data[n,:] = raw[n*len(axis.values):(n+1)*len(axis.values)]
          n+=1
        else:
          mask[i,j] = numpy.nan
    return SparseSpectrumMap(axis, Data2D(data, datatypes.COUNTS), Data3D(mapping, datatypes.POS_um), Data2D(mask, datatypes.NUMBER), None)
  else:
    names = file.readline().strip().split("\t")
    units = file.readline().strip().split("\t")
    length = len(headers)+3
    file.close()
    data = ASCII(filename, skip_header=length).T
    axis = Data1D(data[0,:],DataType(names[0],units[0]))
    if headers["Measurement Type"][0] == "Series":
      #2D
      values = Data2D(data[2:,:],datatypes.COUNTS)
      if headers["Measurement Type"][1] == "PLE Xe Arc Lamp":
        dtype = DataType("Excitation Wavelength","nm")
        if int(headers["Dark Frames"][0])>0:
          #Has been corrected and is offset by 1000
          values.values-=1000
      else:
        dtype = DataType(headers["Measurement Type"][1],units[2])
      keys = Data1D(numpy.array(names[2:2+values.values.shape[0]]).astype(numpy.float),dtype)
      if headers["Measurement Type"][1] == "PLE Xe Arc Lamp":
        return PLE(axis,keys,values,None,None)
      else:
        return SpectrumSeries(axis,keys,values)
    elif headers["Measurement Type"][0] == "PLE Response":
      values = Data1D(data[1,:],DataType(names[1],units[1]))
      return Spectrum(axis, values)
    elif headers["Measurement Type"][0] == "Histogram":
      values = Data1D(data[1,:],datatypes.COUNTS)
      return Spectrum(axis,values)
    else:
      #1D
      values = Data1D(data[2,:],datatypes.COUNTS)
      return Spectrum(axis, values)



import numpy
from fepydas.datatypes.Data import Data1D, Data2D

class Binner:
  def __init__(self):
    return

  def createBins(self, axis, num_bins):
    values  = self.transform(axis.values)
    empty_bins = True
    self.histogram, self.bin_edges = numpy.histogram(values, num_bins)
    self.digitized = numpy.digitize(values, self.bin_edges, right=True) - 1
    print(self.digitized)
    self.bin_centers = (self.bin_edges[:-1]+self.bin_edges[1:])/2
    self.bin_avgs = numpy.zeros_like(self.bin_centers)
    self.bin_stds = numpy.zeros_like(self.bin_centers)
    self.num_bins = num_bins

    for i in range(num_bins):
      idx = numpy.where(self.digitized == i)
      self.bin_avgs[i] = numpy.average(values[idx])
      self.bin_stds[i] = numpy.std(values[idx])

    return Data1D(self.bin_avgs, axis.datatype, self.bin_stds)

  def binData(self, data2D):
    values2D = data2D.values
    originalShape = values2D.shape
    self.binned_data = numpy.ndarray(shape=(len(self.histogram), originalShape[1]), dtype=values2D.dtype)
    self.binned_std = numpy.zeros_like(self.binned_data)
    
    for i in range(self.num_bins):
      idx = numpy.where(self.digitized == i)
      self.binned_data[i,:] = numpy.average(values2D[idx[0],:], axis=0)
      self.binned_std[i,:] = numpy.std(values2D[idx[0],:], axis=0)/numpy.sqrt(self.histogram[i])
    
    #Remove Empty Bins
    non_empty = numpy.where(self.histogram > 0)[0]
    self.histogram = self.histogram[non_empty]
    self.bin_centers = self.bin_centers[non_empty]
    self.bin_avgs = self.bin_avgs[non_empty]
    self.bin_stds = self.bin_stds[non_empty]
    self.binned_data = self.binned_data[non_empty,:]
    self.binned_std = self.binned_std[non_empty,:]
    
    return Data2D(self.binned_data, data2D.datatype, self.binned_std), self.histogram
    


class LinearBinner(Binner):
  def __init__(self):
    super().__init__()

  def transform(self, values):
    return values

  def inverseTransform(self, values):
    return values

class InverseBinner(Binner):
  def __init__(self):
    super().__init__()

  def transform(self, values):
    return 1/values

  def inverseTransform(self, values):
    return 1/values

class LogBinner(Binner):
  def __init__(self):
    super().__init__()

  def transform(self, values):
    return numpy.log10(values)

  def inverseTransform(self, values):
    return numpy.power(10, values)



from fepydas.datatypes.Data import DataType
WL_nm = DataType("Wavelength","nm")
RS_cm = DataType("Raman Shift","cm-1")
COUNTS = DataType("Counts","#")
TIME_s = DataType("Time","s")
TIME_ns = DataType("Time","ns")
TIME_ps = DataType("Time","ps")
E_eV = DataType("Energy","eV")
ANGLE_au = DataType("Angle","a.u.")
NUMBER = DataType("Number","#")
I_mA = DataType("Current","mA")
P_mW = DataType("Power","mW")
POS_um = DataType("Position","µm")
TEMP_K = DataType("Temperature","K")



#!/usr/bin/python3
import numpy
from scipy.constants import speed_of_light,Planck,elementary_charge

class DataType:
  def __init__(self, name: str, unit:str):
    self.name = name
    self.unit = unit

class Data:
  def __init__(self, values: numpy.ndarray, datatype: DataType, errors = None):
    self.values = values
    self.datatype = datatype
    self.errors = errors

  def normalize(self):
    self.values = self.values / numpy.max(self.values)
  
  def divideBy(self, data):
    self.values = self.values/data

  def append(self, dataToAppend):
    self.values = numpy.concatenate((self.values, dataToAppend), axis=0)

  def applyJacobianTransformation(self, axis):
    jacobian = axis.getJacobian()
    self.values = self.values * jacobian

  def linearAntiBloom(self, values, over, width=1, reverse=False):
    max = numpy.max(values)
    length = len(values)
    if reverse:
      for j in range(length-1,0,-1):
        if j-width > 0:
          if values[j-width] > over:
            values[j] = max
    else:
      for j in range(length-1):
        if j+width < length:
          if values[j+width] > over:
            values[j] = max
    return values

class Data1D(Data):
  def __init__(self, values: numpy.ndarray, datatype: DataType, errors = None):
    super().__init__(values,datatype,errors=errors)

  def antiBloom(self, over, width=1, reverse=False):
    self.values = self.linearAntiBloom(self.values, over, width, reverse)

  def cutAndGetIndexRange(self, min, max):
    idx1, idx2 = self.getIndexRange(min,max)
    self.cut(idx1,idx2)
    return idx1, idx2 

  def getIndexRange(self, min, max):
    idx1 = (numpy.abs(self.values - min)).argmin()
    idx2 = (numpy.abs(self.values - max)).argmin()
    if idx2<idx1:
      temp = idx2
      idx2 = idx1
      idx1 = temp
    return idx1,idx2

  def convertToEnergy(self, unit="eV"):
    self.values = WavelengthToEnergy().apply(self.values*1e-9)
    self.datatype = DataType("Energy","eV")

  def getJacobian(self):
    if self.datatype.unit == "eV":
      factor = elementary_charge
    else:
      factor = 1
    return Planck * speed_of_light/ ((factor*self.values)**2)

  def cut(self, idx1, idx2):
    self.values = self.values[idx1:idx2]
    if self.errors is not None:
      self.errors = self.errors[idx1:idx2]
 
  def shiftValues(self, amount):
    self.values += amount

  def trim(self):
    maxIdx = numpy.argmax(self.values)
    zeroes = numpy.where(self.values==0)[0]
    start = 0
    end = zeroes[numpy.where(zeroes > maxIdx)[0][-1]]
    self.values = self.values[start:end]
    self.values = numpy.array(self.values,dtype=numpy.float64)
    self.values /= self.values.sum()
    
        
  def to2D(self, mask):
    newValues = numpy.full_like(mask.values, numpy.nan, dtype=numpy.float)
    x,y = mask.values.shape
    for i in range(x):
      for j in range(y):
        if not numpy.isnan(mask.values[i,j]):
          idx = int(mask.values[i,j])
          newValues[i,j] = self.values[idx]
    return Data2D(newValues, self.datatype)

class Data2D(Data):
  def __init__(self, values: numpy.ndarray, datatype: DataType, errors=None):
    super().__init__(values,datatype,errors=errors)

  def antiBloom(self, over, width=1, reverse=False):
    for i in range(self.values.shape[0]):
      self.values[i,:] = self.linearAntiBloom(self.values[i,:], over, width, reverse)
   
  def average(self):
    return Data1D(numpy.average(self.values,axis=0),self.datatype)

  def collapse(self, filter=None):
    if filter:
      return Data1D(numpy.sum(self.values[filter],axis=0), self.datatype)
    else:
      return Data1D(numpy.sum(self.values,axis=0),self.datatype)

  def integrate(self):
    if self.errors is None:
      errors = None
    else:
      errors = numpy.sum(self.errors, axis=1)
    return Data1D(numpy.sum(self.values,axis=1),self.datatype, errors)

  def normalizeIndividual(self):
    maxima = numpy.max(self.values,axis=1)
    self.values = (self.values.T / maxima).T

  def cut(self, idx1, idx2):
    self.values = self.values[:,idx1:idx2]
    if self.errors is not None:
      self.errors = self.errors[:,idx1:idx2]
      
  def cutMap(self, idx1, idx2, idx3, idx4):
    self.values = self.values[idx1:idx2, idx3:idx4]

  def getCrosssectionAt(self, idx):
    return Data1D(self.values[:,idx], self.datatype)
 
  

class Data3D(Data):
  def __init__(self, values: numpy.ndarray, datatype: DataType):
    super().__init__(values,datatype)

  def collapse(self, filter=None):
    print(filter)
    if filter is not None:
      return Data1D(numpy.sum(self.flatten()[filter],axis=0), self.datatype)
    else:
      return Data1D(numpy.sum(self.flatten(),axis=0),self.datatype)

  def cutMap(self, xMin, xMax, yMin, yMax):
    self.values = self.values[xMin:xMax,yMin:yMax,:]

  def flatten(self, filter=None):
    if filter is None:
      x,y,z = self.values.shape
      return self.values.reshape(x*y,z)
    else:
      return self.values[filter].reshape(len(filter[0]),self.values.shape[2])


  def getIndexRange(self, xMin, xMax, yMin, yMax):
    idx1 = (numpy.abs(self.values[:,0,0] - xMin)).argmin(axis=0)
    idx2 = (numpy.abs(self.values[:,0,0] - xMax)).argmin(axis=0)
    idx3 = (numpy.abs(self.values[0,:,1] - yMin)).argmin(axis=0)
    idx4 = (numpy.abs(self.values[0,:,1] - yMax)).argmin(axis=0)  
    return idx1,idx2,idx3,idx4

  def get1DAt(self,x,y):
    return Data1D(self.values[x,y,:], self.datatype)

  def cutAndGetIndexRange(self, xMin, xMax, yMin, yMax):
    i1,i2,i3,i4 = self.getIndexRange(xMin,xMax,yMin,yMax)
    self.cutMap(i1,i2,i3,i4)
    return i1,i2,i3,i4

  def integrate(self):
    return Data2D(numpy.sum(self.values, axis=2), self.datatype)

  def average(self):
    return Data1D(numpy.average(self.values, axis=(0,1)), self.datatype)

  def extractAxes(self):
    y = self.values[0,:,1]
    x = self.values[:,0,0]
    return Data1D(x, self.datatype), Data1D(y, self.datatype)
  
  def cut(self, idx1, idx2):
    self.values = self.values[:,:,idx1:idx2]

  def normalizeIndividual(self):
    maxima = numpy.max(self.values,axis=2)
    self.values = (self.values.T / maxima.T).T

class DataTransformation():
  def __init__(self, transformation, newUnit):
    self.transformation = transformation
    self.newUnit = newUnit

  def apply(self, oldAxis):
    newValues = self.transformation.apply(oldAxis.values)
    if oldAxis.errors is not None:
      errors = oldAxis.errors / oldAxis.values * newValues
    else:
      errors = None
    return Data1D(self.transformation.apply(oldAxis.values), self.newUnit, errors)

class ArrheniusTransformation(DataTransformation):
  def __init__(self):
    super().__init__(Transformation(self.reciprocal,{}),DataType("1/T","1/K"))

  def reciprocal(self, values):
    return 1/values
    
class Transformation():
  def __init__(self, function, parameterDict):
    self.function = function
    self.params = parameterDict

  def apply(self, values):
    if values is None: return None
    return self.function(values,**self.params)

class WavelengthToEnergy(Transformation):
  def __init__(self, unit="eV"):
    params = {}
    if unit=="eV":
      params["factor"]=1/elementary_charge
    else:
      params["factor"]=1
    super().__init__(self.wavelengthToEnergy, params)

  def wavelengthToEnergy(self, values, factor):
    return factor * Planck * speed_of_light / values
    
class VacuumCorrection(Transformation):
  def __init__(self, temperature=20, pressure=101325, humidity=0, co2=610):
    params = {}
    params["temperature"]=temperature
    params["pressure"]=pressure
    params["humidity"]=humidity
    params["co2"]=co2
    super().__init__(self.convertWavelength, params)
    
  def convertWavelength(self, values, temperature, pressure, humidity, co2):
    return values * self.n(values/1000, temperature, pressure, humidity, co2)
  
  def Z(self, T,p,xw): #compressibility
    t=T-273.15
    a0 = 1.58123e-6   #K·Pa^-1
    a1 = -2.9331e-8   #Pa^-1
    a2 = 1.1043e-10   #K^-1·Pa^-1
    b0 = 5.707e-6     #K·Pa^-1
    b1 = -2.051e-8    #Pa^-1
    c0 = 1.9898e-4    #K·Pa^-1
    c1 = -2.376e-6    #Pa^-1
    d  = 1.83e-11     #K^2·Pa^-2
    e  = -0.765e-8    #K^2·Pa^-2
    return 1-(p/T)*(a0+a1*t+a2*t**2+(b0+b1*t)*xw+(c0+c1*t)*xw**2) + (p/T)**2*(d+e*xw**2)
  
  def n(self, λ,t,p,h,xc):
    # λ: wavelength, 0.3 to 1.69 μm 
    # t: temperature, -40 to +100 °C
    # p: pressure, 80000 to 120000 Pa
    # h: fractional humidity, 0 to 1
    # xc: CO2 concentration, 0 to 2000 ppm
    σ = 1/λ           #μm^-1
    T= t + 273.15     #Temperature °C -> K
    R = 8.314510      #gas constant, J/(mol·K)
    k0 = 238.0185     #μm^-2
    k1 = 5792105      #μm^-2
    k2 = 57.362       #μm^-2
    k3 = 167917       #μm^-2
    w0 = 295.235      #μm^-2
    w1 = 2.6422       #μm^-2
    w2 = -0.032380    #μm^-4
    w3 = 0.004028     #μm^-6
    A = 1.2378847e-5  #K^-2
    B = -1.9121316e-2 #K^-1
    C = 33.93711047
    D = -6.3431645e3  #K
    α = 1.00062
    β = 3.14e-8       #Pa^-1,
    γ = 5.6e-7        #°C^-2
    #saturation vapor pressure of water vapor in air at temperature T
    if(t>=0):
        svp = numpy.exp(A*T**2 + B*T + C + D/T) #Pa
    else:
        svp = 10**(-2663.5/T+12.537)
    #enhancement factor of water vapor in air
    f = α + β*p + γ*t**2
    #molar fraction of water vapor in moist air
    xw = f*h*svp/p
    #refractive index of standard air at 15 °C, 101325 Pa, 0% humidity, 450 ppm CO2
    nas = 1 + (k1/(k0-σ**2)+k3/(k2-σ**2))*1e-8
    #refractive index of standard air at 15 °C, 101325 Pa, 0% humidity, xc ppm CO2
    naxs = 1 + (nas-1) * (1+0.534e-6*(xc-450))
    #refractive index of water vapor at standard conditions (20 °C, 1333 Pa)
    nws = 1 + 1.022*(w0+w1*σ**2+w2*σ**4+w3*σ**6)*1e-8
    Ma = 1e-3*(28.9635 + 12.011e-6*(xc-400)) #molar mass of dry air, kg/mol
    Mw = 0.018015                            #molar mass of water vapor, kg/mol
    Za = self.Z(288.15, 101325, 0)                #compressibility of dry air
    Zw = self.Z(293.15, 1333, 1)                  #compressibility of pure water vapor
    #Eq.4 with (T,P,xw) = (288.15, 101325, 0)
    ρaxs = 101325*Ma/(Za*R*288.15)           #density of standard air
    #Eq 4 with (T,P,xw) = (293.15, 1333, 1)
    ρws  = 1333*Mw/(Zw*R*293.15)             #density of standard water vapor
    # two parts of Eq.4: ρ=ρa+ρw
    ρa   = p*Ma/(self.Z(T,p,xw)*R*T)*(1-xw)       #density of the dry component of the moist air    
    ρw   = p*Mw/(self.Z(T,p,xw)*R*T)*xw           #density of the water vapor component
    nprop = 1 + (ρa/ρaxs)*(naxs-1) + (ρw/ρws)*(nws-1)
    return nprop


class Response():
  def __init__(self, input: Data1D, values: Data1D):
    self.input = input
    self.values = values 

  def interpolateResponse(self, inputValues: numpy.ndarray, zeroFloor=False):
    if zeroFloor:
      self.values.values -= numpy.amin(self.values.values)*0.9
    return numpy.interp(inputValues, self.input.values, self.values.values)



#!/usr/bin/python3
import numpy
import os
from fepydas.datatypes.Data import Data1D,Data2D,Data3D,Response
from fepydas.datatypes.Dataset import Spectrum,SpectrumSeries,PLE,SpectrumMap
import fepydas.datatypes as datatypes
from fepydas.constructors.datareaders.Generic import ASCII, ASCII_SpectrumRowSeries

def ASCII_RamanSeries(folder):
  files = [f for f in os.listdir(folder) if f.endswith(".txt")]
  num = len(files)
  ref = ASCII(os.path.join(folder,files[0])).T
  
  shape = (num, ref.shape[1])
  xAxis = Data1D(ref[0,:],datatypes.RS_cm)
  vals = numpy.ndarray(shape=shape)
  for i,file in enumerate(sorted(files)):
    spec = ASCII(os.path.join(folder,file)).T
    vals[i,:] = spec[1,:]

  values = Data2D(vals,datatypes.COUNTS)
  keys = Data1D(numpy.arange(num),datatypes.ANGLE_au)
  return SpectrumSeries(xAxis,keys,values)

  #data = ASCII(filename).T
  #ramanType = DataType("Raman Shift", "cm-1")
  #countType = DataType("Counts","#")
  #xAxis = Data1D

def ASCII_Spectrum(filename):
  data = ASCII(filename,skip_header=1).T
  xAxis = Data1D(data[0,:],datatypes.WL_nm)
  yAxis = Data1D(data[2,:],datatypes.COUNTS)
  return Spectrum(xAxis,yAxis)

def ASCII_TimeSeries(filename):
  data = ASCII(filename).T
  xAxis = Data1D(data[0,1:],datatypes.WL_nm)
  vals = Data2D(data[1:,1:],datatypes.COUNTS)
  keys = Data1D(data[1:,0],datatypes.TIME_s)
  return SpectrumSeries(xAxis,keys,vals)

def ASCII_Gelbes(filename):
  data = ASCII(filename,skip_header=3).T
  xAxis = Data1D(data[0,:],datatypes.WL_nm)
  vals = Data1D(data[1,:],datatypes.COUNTS)
  return Spectrum(xAxis,vals)

def ASCII_TempSeries_Gelbes(filename):
  data = ASCII(filename, skip_header=3).T
  xAxis = Data1D(data[0,1:],datatypes.WL_nm)
  vals = Data2D(data[1:,1:],datatypes.COUNTS)
  keys = Data1D(data[1:,0],datatypes.TEMP_K)
  return SpectrumSeries(xAxis,keys,vals)

def ASCII_TimeSeries_Gelbes(filename):
  data = ASCII(filename,skip_header=3).T
  xAxis = Data1D(data[0,:],datatypes.WL_nm)
  vals = Data2D(data[1:,:],datatypes.COUNTS)
  keys = Data1D(numpy.arange(vals.values.shape[0]),datatypes.TIME_s)
  return SpectrumSeries(xAxis,keys,vals)

def ASCII_SSTRPL(filename):
  return ASCII_SpectrumRowSeries(filename, datatypes.TIME_ps, datatypes.NUMBER, datatypes.COUNTS)

def ASCII_EQE(filename):
  data = ASCII(filename)
  xAxis = Data1D(data[0,3:],datatypes.WL_nm)
  keyAxis = Data1D(data[2:,1]-data[1,1],datatypes.I_mA)
  vals = Data2D(data[2:,3:]-data[1,3:],datatypes.COUNTS)
  return SpectrumSeries(xAxis,keyAxis,vals)

def ASCII_PLE(PLEFile, CalibrationFile=None, ResponseFile=None):
  if CalibrationFile:
    data = ASCII(CalibrationFile,skip_header=3).T
    calX = Data1D(data[0,:],datatypes.WL_nm)
    calY = Data1D(data[1,:],datatypes.COUNTS)
  else:
    calY = None

  data = ASCII(PLEFile,skip_header=3).T
  pleX = Data1D(data[0,1:],datatypes.WL_nm)
  pleKeys = Data1D(data[1:,0],datatypes.WL_nm)
  pleVals = Data2D(data[1:,1:],datatypes.COUNTS)

  if CalibrationFile:
    if numpy.sum(pleX.values-calX.values)>0:
      print("Warning, PLE and Calibration Spectral Axes not compatible!!")

  data = ASCII(ResponseFile,skip_header=1).T
  responseX = Data1D(data[0,:],datatypes.WL_nm)
  responseY = Data1D(data[1,:],datatypes.P_mW)
  
  return PLE(pleX, pleKeys, pleVals, calY, Response(responseX,responseY))


def MapscanCommon(filename, ininame,   format, CalibrationFile=None):



    if CalibrationFile:
      data = ASCII(CalibrationFile,skip_header=3).T
      calX = Data1D(data[0,:],datatypes.WL_nm)
      calY = Data1D(data[1,:],datatypes.COUNTS)
    else:
      calY = None

    data  =  numpy.fromfile(filename,dtype=format)
    ini = open(ininame)
    inilines = ini.readlines()
    ini.close()
    xDim, yDim, xAmp, yAmp, center, dispersion = 0, 0, 0, 0, 0, 0
    for line in inilines:
        line=line.strip()
        print(line)
        line = line.split("=")
        if len(line) > 1:
            if(line[0].startswith("x-dim")):
                xDim = int(line[1])
            if(line[0].startswith("y-dim")):
                yDim = int(line[1])
            if(line[0].startswith("x-amplitude")):
                xAmp = float(line[1].replace(",","."))
            if(line[0].startswith("y-amplitude")):
                yAmp = float(line[1].replace(",","."))
            if(line[0].startswith("center")):
                center = float(line[1].replace(",","."))
            if(line[0].startswith("dispersion")):
                dispersion = float(line[1].replace(",","."))
    if xDim == 0 or yDim == 0 or xAmp == 0 or yAmp == 0 or center == 0 or dispersion == 0:
        print("Error: some parameters are missing in ini file")
    print("Coordinates: {0}x{1} Size: {2}x{3}".format(xDim,yDim,xAmp,yAmp))
    print("Center: {0} Dispersion: {1}".format(center, dispersion))


    xData = numpy.arange(center+1023.5*dispersion,  center-1024.5*dispersion,  -dispersion)[:2048]
    
    axis = Data1D(xData, datatypes.WL_nm)
    
    yData = numpy.ndarray(shape=(xDim, yDim,  2048))
    
    mapping = numpy.ndarray(shape=(xDim, yDim, 2),dtype=numpy.float)
    
    for x in range(xDim):
        for y in range(yDim):
            mapping[x,y,0]=x/xDim*xAmp
            mapping[x,y,1]=y/yDim*yAmp
            yData[x,y]=data[2048*(x*yDim+y):2048*(x*yDim+y+1)]

    data = Data3D(yData, datatypes.COUNTS)
    mapping = Data3D(mapping, datatypes.POS_um)
    
    return SpectrumMap(axis, data, mapping, calY)
    
def Mapscan(filename,  calibration=None):
    return MapscanCommon(filename, filename+"ini", numpy.dtype(">f8"), calibration)

def Mapscan2(filename, calibration=None):
    return MapscanCommon(filename,filename[:len(filename)-1]+"ini", numpy.dtype(">u2"), calibration)



import numpy
from lmfit import Parameters
def Lorentzian(x, bg=0, I=1, x0=0, fwhm=1):
  if I==0:
    return 0
  t = (x-x0)/(fwhm/2)
  return bg + I / (1+t*t)

def Gaussian(x, bg=0, I=1, x0=0, fwhm=1):
  if I==0:
    return 0
  return I * numpy.exp(- numpy.log(2)*((x-x0)/(fwhm/2))**2)

def Linear(x, a=1, b=0):
  return a*x+b

def Quadratic(x, a=0, b=1, c=0):
  return a*x**2+b*x+c

def Exponential(x, bg=0 ,I=1, x0=0, tau=1, rise=0.001):
  y = numpy.zeros_like(x)
  y+=bg
  decay = numpy.where(x>=x0)

  y[decay]+= I*numpy.exp(-(x[decay]-x0)/tau)*(1-numpy.exp(-(x[decay]-x0)/rise))
#  notDecay = numpy.where(x<x0)
#  y[decay]+= I*numpy.exp(-(x[decay]-x0)/tau)
#  y[notDecay]+= I*numpy.exp(-(x0-x[notDecay])/rise)
  return y


def BiExponential(x, bg=0 , x0=0, I_1=1, tau_1=1, I_2=1, tau_2=1, rise=0.001):
  y = numpy.zeros_like(x)
  y+=bg
  decay = numpy.where(x>=x0)
  y[decay]+= (I_1*numpy.exp(-(x[decay]-x0)/tau_1)+I_2*numpy.exp(-(x[decay]-x0)/tau_2))*(1-numpy.exp(-(x[decay]-x0)/rise))
  #notDecay = numpy.where(x<x0)
  #y[decay]+= I_1*numpy.exp(-(x[decay]-x0)/tau_1)+I_2*numpy.exp(-(x[decay]-x0)/tau_2)
  #y[notDecay]+= (I_1+I_2)*numpy.exp(-(x0-x[notDecay])/rise)
  return y

def TriExponential(x, bg=0 , x0=0, I_1=1, tau_1=1, I_2=1, tau_2=1, I_3=1, tau_3=1, rise=0.001):
  y = numpy.zeros_like(x)
  y+=bg
  decay = numpy.where(x>=x0)
  notDecay = numpy.where(x<x0)
  y[decay]+= I_1*numpy.exp(-(x[decay]-x0)/tau_1)+I_2*numpy.exp(-(x[decay]-x0)/tau_2)+I_3*numpy.exp(-(x[decay]-x0)/tau_3)
  y[notDecay]+= (I_1+I_2+I_3)*numpy.exp(-(x0-x[notDecay])/rise)
  return y


def BiExponentialTail(x, bg=0, I_1=1, tau_1=1, I_2=1, tau_2=1):
  y = bg + I_1*numpy.exp(-(x)/tau_1)+I_2*numpy.exp(-(x)/tau_2)
  return y



#!/usr/bin/python3
import numpy
from pylab import *
from matplotlib.colors import *
from fepydas.datatypes.Data import Data,Data1D,Data2D, Data3D
from fepydas.datatypes.Dataset import Spectrum,SpectrumWithIRF,SpectrumSeries, Map
from fepydas.workers.Fit import Fit

class BasePlot:
  def __init__(self, polar=False):
    self.figure = figure()
    self.bottom = self.figure.add_subplot(111, polar=polar)
  
  def legend(self):
    self.bottom.legend()

  def show(self):
    show()

  def save(self, filename):
    self.figure.savefig(filename)

  def setXAxis(self, axis: Data1D):
    self.bottom.set_xlabel("{0} ({1})".format(axis.datatype.name, axis.datatype.unit))
 
  def setXRange(self, xmin, xmax):
    self.bottom.set_xlim(xmin,xmax)
 
  def setYRange(self, ymin, ymax):
    self.bottom.set_ylim(ymin,ymax)

  def setXLog(self):
    self.bottom.set_xscale("log")

  def setYLog(self):
    self.bottom.set_yscale("log")

  def setYAxis(self, data: Data):
    self.bottom.set_ylabel("{0} ({1})".format(data.datatype.name, data.datatype.unit))

class Plot1D(BasePlot):
  def __init__(self, axis: Data1D, polar=False):
    super().__init__(polar=polar)
    if polar:
      axis.values = axis.values/360*2*numpy.pi
    self.setXAxis(axis)
  def plotLine(self, x, y, label=None, linewidth=2):
    self.bottom.plot(x, y, label=label, linewidth=linewidth)
  def plotPoints(self, x, y, xerr=None, yerr=None, label=None):
    self.bottom.errorbar(x, y, xerr=xerr, yerr=yerr, label=label, fmt="o", alpha=0.2)

class Plot1DSingleAxis(Plot1D):
  def __init__(self, axis: Data1D, polar=False):
    super().__init__(axis, polar=polar)
    self.axis = axis
  def addLine(self, data: ndarray, label=None, linewidth=2):
    #if len(self.axis.values) > len(data):
    self.plotLine(self.axis.values[0:len(data)], data, label=label, linewidth=linewidth)
  def addPoints(self, data: ndarray, errors = None, label=None):
    self.plotPoints(self.axis.values, data, self.axis.errors, errors, label=label)
  def addLines(self, data, labels):
    for i in range(data.shape[0]):
      self.plotLine(self.axis.values, data[i], label=labels[i])
    
class Plot1DMultiAxis(Plot1D):
  def __init__(self, axis: Data1D, data: Data1D):
    super().__init__(axis)
    self.setYAxis(data)
  def addLine(self, axis: Data1D, data: Data1D, label=None):
    #TODO: Compare Units
    self.plotLine(axis.values, data.values, label=label)

  def addPoints(self, axis: Data1D, data: Data1D, label=None):
    #TODO: Compare Units
    self.plotPoints(axis.values, data.values, label=label)

class MultiSpectrumScatterPlot(Plot1DMultiAxis):
  def __init__(self, spectra, labels):
    super().__init__(spectra[0].axis,spectra[0].data)
    for i, spectrum in enumerate(spectra):
      self.addPoints(spectrum.axis,spectrum.data,labels[i])

class Plot1DSingleLine(Plot1DSingleAxis):
  def __init__(self, axis: Data1D, data: Data1D, label=None):
    super().__init__(axis)
    self.setYAxis(data)
    self.addLine(data.values, label=label)

class Plot1DSingleScatter(Plot1DSingleAxis):
  def __init__(self, axis: Data1D, data: Data1D):
    super().__init__(axis)
    self.setYAxis(data)
    self.addPoints(data.values, data.errors)

class Plot1DMultiLineSingleAxis(Plot1DSingleAxis):
  def __init__(self, axis: Data1D, keys: Data1D, data: Data2D, polar=False):
    super().__init__(axis, polar=polar)
    self.setYAxis(data)
    for i,key in enumerate(keys.values):
      self.addLine(data.values[i], key)

class Plot1DMultiScatterSingleAxis(Plot1DSingleAxis):
  def __init__(self, axis: Data1D, keys: Data1D, data: Data2D):
    super().__init__(axis)
    self.setYAxis(data)
    for i,key in enumerate(keys.values):
      self.addPoints(data.values[i], key)

class SpectrumPlot(Plot1DSingleLine):
  def __init__(self, spectrum: Spectrum, label = None):
    super().__init__(spectrum.axis, spectrum.data, label=label)
  def addSpectrum(self, spectrum, label=None):
    self.addLine(spectrum.data.values, label=label)

class HistoPlot(Plot1D):
  def __init__(self, spectrum: Spectrum, label=None, n=100, bins=None):
    super().__init__(spectrum.data) #In a histogram, the data values form the axis
    if bins is not None:
      n, bins, patches = hist(spectrum.data.values.flatten(),bins=bins)
    else:
      n, bins, patches = hist(spectrum.data.values.flatten(),n)
    self.data = n

class SpectrumScatterPlot(Plot1DSingleScatter):
  def __init__(self, spectrum: Spectrum):
    super().__init__(spectrum.axis, spectrum.data)

class SpectrumScatterPlotWithFit(SpectrumScatterPlot):
  def __init__(self, spectrum: Spectrum, fit: Fit):
    super().__init__(spectrum)
    self.addLine(fit.evaluate(spectrum.axis.values))

class SpectrumWithIRFPlot(Plot1DSingleScatter):
  def __init__(self, spectrum: SpectrumWithIRF):
    super().__init__(spectrum.axis, spectrum.data)
    self.addLine(spectrum.getExtendedIRF())

class SpectrumSeriesPlot(Plot1DMultiLineSingleAxis):
  def __init__(self, spectra: SpectrumSeries, polar=False):
    super().__init__(spectra.axis, spectra.keys, spectra.data, polar=polar)

class SpectrumSeriesScatterPlot(Plot1DMultiScatterSingleAxis):
  def __init__(self, spectra):
    super().__init__(spectra.axis, spectra.keys, spectra.data)

class Plot2D(BasePlot):
  def __init__(self, axis, keys):
    super().__init__()
    self.setXAxis(axis)
    self.setYAxis(keys)


  def setZAxis(self, data: Data):
    self.colorbar.set_label("{0} ({1})".format(data.datatype.name, data.datatype.unit))
    self.colorbar.minorticks_on()

class ContourPlot(Plot2D):
  def __init__(self, spectra: SpectrumSeries, log = False):
    super().__init__(spectra.axis, spectra.keys)
    cmap = LinearSegmentedColormap.from_list('my_colormap',["black","red","green","blue","white"],1000)
    extent = [spectra.axis.values[0],spectra.axis.values[-1],spectra.keys.values[0],spectra.keys.values[-1]]
    if log:
      values = numpy.log10(spectra.data.values+0.001)
    else:
      values = spectra.data.values
    image = imshow(values,interpolation="none",cmap=cmap,origin="lower",aspect="auto",extent=extent)
    
class ContourMapPlot(Plot2D):
  def __init__(self, map: Map, log = False, colors=["black","red","orange","yellow","green","blue","white"], zRange=None):
    x, y = map.mapping.extractAxes()
    super().__init__(x, y)
    cmap = LinearSegmentedColormap.from_list('my_colormap',colors,1000)
    cmap.set_over("white")
    cmap.set_under("black")
    cmap.set_bad("violet")
    extent = [x.values[0],x.values[-1],y.values[0],y.values[-1]]
    if log:
      values = numpy.log10(map.data.values.T+0.001)
    else:
      values = map.data.values.T
    if zRange:
      zmin=zRange[0]
      zmax=zRange[1]
    else:
      zmin=None
      zmax=None
    image = imshow(values,interpolation="none",cmap=cmap,origin="lower",aspect="auto",extent=extent,vmin=zmin,vmax=zmax)
    self.colorbar = self.figure.colorbar(image, cmap=cmap)#,  ticks=range(len(colors)))
    self.setZAxis(map.data)
    



import os

class DirectoryLoader:
  def __init__(self, constructor):
    self.constructor = constructor
    self.loaded = {}

  def loadFromDirectory(self, directory, filter_function=None):
    files = [f for f in os.listdir(directory) if os.path.isfile(os.path.join(directory,f))]
    print(files)
    for filename in sorted(files):
      if filter_function is None or filter_function(filename):   
        self.loaded[filename] = self.constructor(os.path.join(directory,filename))

  def callFunction(self, function, *args):
    for key in self.loaded.keys():
      getattr(self.loaded[key], function)(*args)

  def getList(self):
    return list(self.loaded.values())

  def getKeys(self):
    return list(self.loaded.keys())



# -*- coding: utf-8 -*-
# sdtfile.py

# Copyright (c) 2007-2013, Christoph Gohlke
# Copyright (c) 2007-2013, The Regents of the University of California
# Produced at the Laboratory for Fluorescence Dynamics
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the copyright holders nor the names of any
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Read Becker & Hickl SDT files.

SDT files are produced by Becker & Hickl SPCM software. They contain time
correlated single photon counting instrumentation parameters and measurement
data. Currently only the "Setup & Data" and "FCS Data" formats are supported.

`Becker & Hickl GmbH <http://www.becker-hickl.de/>`_ is a manufacturer of
equipment for photon counting.

:Author:
  `Christoph Gohlke <http://www.lfd.uci.edu/~gohlke/>`_

:Organization:
  Laboratory for Fluorescence Dynamics, University of California, Irvine

:Version: 2013.01.18

Requirements
------------
* `CPython 2.7, 3.2 or 3.3 <http://www.python.org>`_
* `Numpy 1.7 <http://www.numpy.org>`_

References
----------
(1) W Becker. The bh TCSPC Handbook. Third Edition. Becker & Hickl GmbH 2008.
    pp 401.
(2) SPC_data_file_structure.h header file. Part of the Becker & Hickl
    SPCM software.

Examples
--------
>>> sdt = SdtFile("image.sdt")
>>> sdt.header.revision
588
>>> sdt.info.id[1:-1]
b'SPC Setup & Data File'
>>> int(sdt.measure_info.scan_x)
128
>>> sdt.data[0].shape
(128, 128, 256)
>>> sdt.times[0].shape
(256,)

>>> sdt = SdtFile("fluorescein.sdt")
>>> len(sdt.data)
4
>>> sdt.data[3].shape
(1, 1024)

>>> sdt = SdtFile("fcs.sdt")
>>> sdt.info.id[1:-1]
b'SPC FCS Data File'
>>> sdt.data[0].shape
(512, 512, 256)

"""

from __future__ import division, print_function

import sys

import numpy

__version__ = '2013.01.18'
__docformat__ = 'restructuredtext en'
__all__ = ["SdtFile"]


class SdtFile(object):
    """Becker & Hickl SDT file.

    Attributes
    ----------
    header : numpy.rec.array of FILE_HEADER structure
        General information about the location of the setup and measurement
        data within the file.
    info : FileInfo
        General information in ASCII format.
    setup : SetupBlock
        Setup block containing all system parameters, display parameters, etc.
    measure_info : numpy.rec.array of MEASURE_INFO structure
        Measurement description blocks.
    block_headers : list of numpy.rec.array of BLOCK_HEADER structure
        Data block headers.
    data : list of 2D numpy arrays
        Photon counts at each curve point.
    times : list of 1D numpy arrays
        Time axes for each data set.

    """
    def __init__(self, arg=None):
        """Initialize instance from filename or open file."""
        if isinstance(arg, basestring):
            with open(arg, 'rb') as fh:
                self._fromfile(fh)
        elif hasattr(arg, 'seek'):
            self._fromfile(arg)

    def _fromfile(self, fh):
        """Initialize instance from open file."""
        # read file header
        self.header = numpy.rec.fromfile(fh, dtype=FILE_HEADER,
                                         shape=1, byteorder='<')[0]
        if self.header.header_valid != 0x5555:
            raise ValueError("not a SDT file")
        if self.header.no_of_data_blocks == 0x7fff:
            self.header.no_of_data_blocks = self.header.reserved1
        elif self.header.no_of_data_blocks > 0x7fff:
            raise ValueError("")

        # read file info
        fh.seek(self.header.info_offset)
        self.info = FileInfo(fh.read(self.header.info_length))
        try:
            if self.info.id not in (b"SPC Setup & Data File",
                                    b"SPC FCS Data File"):
                raise NotImplementedError(
                    "currently not supported:", self.info.id)
        except AttributeError:
            raise ValueError("invalid SDT file info\n", self.info)

        # read setup block
        fh.seek(self.header.setup_offs)
        self.setup = SetupBlock(fh.read(self.header.setup_length))

        # read measurement description blocks
        fh.seek(self.header.meas_desc_block_offset)
        self.measure_info = numpy.rec.fromfile(
            fh, dtype=MEASURE_INFO, shape=self.header.no_of_meas_desc_blocks,
            byteorder='<')

        self.times = []
        self.data = []
        self.block_headers = []
        offset = self.header.data_block_offset
        for _ in range(self.header.no_of_data_blocks):
            # read data block header
            fh.seek(offset)
            bh = numpy.rec.fromfile(fh, dtype=BLOCK_HEADER, shape=1,byteorder='<')[0]
            self.block_headers.append(bh)
            # read data block
            mi = self.measure_info[bh.meas_desc_block_no]
            dtype = BlockType(bh.block_type).dtype
            dsize = bh.block_length // dtype.itemsize
            data = numpy.fromfile(fh, dtype=dtype, count=dsize)
            if dsize == mi.scan_x * mi.scan_y * mi.adc_re:
                data = data.reshape(mi.scan_x, mi.scan_y, mi.adc_re)
            elif dsize == mi.image_x * mi.image_y * mi.adc_re:
                data = data.reshape(mi.image_x, mi.image_y, mi.adc_re)
            else:
                data = data.reshape(-1, mi.adc_re)
            self.data.append(data)
            # generate time axis
            t = numpy.arange(mi.adc_re, dtype=numpy.float64)
            r = (mi.tac_r / mi.tac_g) / mi.adc_re
            t *= r
            self.times.append(t)
            offset = bh.next_block_offs

    def block_measure_info(self, block):
        """Return measure_info record for data block."""
        return self.measure_info[self.data_headers[block].meas_desc_block_no]

    def __str__(self):
        """Return string containing all information about SDT file."""
        return "\n\n".join([str(i) for i in (
            self.name, self.header, self.info, self.measure_info,
            self.data_headers, self.data[0].shape)])


class FileInfo(str):
    """File info string and attributes."""
    def __init__(self, value):
        assert (value.startswith(b'*IDENTIFICATION') and
                value.strip().endswith(b'*END'))
        for line in value.splitlines()[1:-1]:
            try:
                key, value = line.split(b':', 1)
            except Exception:
                pass
            else:
                setattr(self, bytes2str(key.strip().lower()), value.strip())


class SetupBlock(object):
    """Setup block ascii and binary data."""
    def __init__(self, value):
        assert (value.startswith(b'*SETUP') and
                value.strip().endswith(b'*END'))
        i = value.find(b'BIN_PARA_BEGIN')
        if i:
            self.ascii = value[:i]
            self.binary = bytes(value[i+15:-10])
            # todo: parse binary data here
        else:
            self.ascii = value
            self.binary = None

    def __str__(self):
        return self.ascii


class BlockNo(object):
    """The lblock_no field of BLOCK_HEADER."""
    def __init__(self, value):
        self.data = (value & 0xFFFFFF00) >> 24
        self.module = value & 0x000000FF

    def __str__(self):
        return "Data number: %s\nModule number: %s" % (self.data, self.module)

    def __iter__(self):
        return iter((self.data, self.module))


class BlockType(object):
    """The block_type field of BLOCK_HEADER."""
    def __init__(self, value):
        self.mode = BLOCK_CREATION[value & 0xF]
        self.contents = BLOCK_CONTENT[value & 0xF0]
        self.dtype = BLOCK_DTYPE[value & 0xF00]

    def __str__(self):
        return "Mode: %s\nContent: %s\nData Type: %s" % (
            self.mode, self.contents, self.dtype)

    def __iter__(self):
        return iter((self.mode, self.contents, self.dtype))


FILE_HEADER = [
    ('revision', 'i2'),
    ('info_offset', 'i4'),
    ('info_length', 'i2'),
    ('setup_offs', 'i4'),
    ('setup_length', 'i2'),
    ('data_block_offset', 'i4'),
    ('no_of_data_blocks', 'i2'),
    ('data_block_length', 'i4'),
    ('meas_desc_block_offset', 'i4'),
    ('no_of_meas_desc_blocks', 'i2'),
    ('meas_desc_block_length', 'i2'),
    ('header_valid', 'u2'),
    ('reserved1', 'u4'),
    ('reserved2', 'u2'),
    ('chksum', 'u2')]

SETUP_BIN_HDR = [
    ('soft_rev', 'u4'),
    ('para_length', 'u4'),
    ('reserved1', 'u4'),
    ('reserved2', 'u2')]

MEASURE_STOP_INFO = [  # Info collected when measurement finished
    ('status', 'u2'),
    ('flags', 'u2'),
    ('stop_time', 'f4'),
    ('cur_step', 'i4'),
    ('cur_cycle', 'i4'),
    ('cur_page', 'i4'),
    ('min_sync_rate', 'f4'),
    ('min_cfd_rate', 'f4'),
    ('min_tac_rate', 'f4'),
    ('min_adc_rate', 'f4'),
    ('max_sync_rate', 'f4'),
    ('max_cfd_rate', 'f4'),
    ('max_tac_rate', 'f4'),
    ('max_adc_rate', 'f4'),
    ('reserved1', 'i4'),
    ('reserved2', 'f4')]

MEASURE_FCS_INFO = [  # Info collected when FIFO measurement finished
    ('chan', 'u2'),
    ('fcs_decay_calc', 'u2'),
    ('mt_resol', 'u4'),
    ('cortime', 'f4'),
    ('calc_photons', 'u4'),
    ('fcs_points', 'i4'),
    ('end_time', 'f4'),
    ('overruns', 'u2'),
    ('fcs_type', 'u2'),
    ('cross_chan', 'u2'),
    ('mod', 'u2'),
    ('cross_mod', 'u2'),
    ('cross_mt_resol', 'u4')]

HIST_INFO = [  # Extension of MeasFCSInfo for other histograms
    ('fida_time', 'f4'),
    ('filda_time', 'f4'),
    ('fida_points', 'i4'),
    ('filda_points', 'i4'),
    ('mcs_time', 'f4'),
    ('mcs_points', 'i4')]

MEASURE_INFO = [  # Measurement description blocks
    ('time', 'a9'),
    ('date', 'a11'),
    ('mod_ser_no', 'a16'),
    ('meas_mode', 'i2'),
    ('cfd_ll', 'f4'),
    ('cfd_lh', 'f4'),
    ('cfd_zc', 'f4'),
    ('cfd_hf', 'f4'),
    ('syn_zc', 'f4'),
    ('syn_fd', 'i2'),
    ('syn_hf', 'f4'),
    ('tac_r', 'f4'),
    ('tac_g', 'i2'),
    ('tac_of', 'f4'),
    ('tac_ll', 'f4'),
    ('tac_lh', 'f4'),
    ('adc_re', 'i2'),
    ('eal_de', 'i2'),
    ('ncx', 'i2'),
    ('ncy', 'i2'),
    ('page', 'u2'),
    ('col_t', 'f4'),
    ('rep_t', 'f4'),
    ('stopt', 'i2'),
    ('overfl', 'u1'),
    ('use_motor', 'i2'),
    ('steps', 'u2'),
    ('offset', 'f4'),
    ('dither', 'i2'),
    ('incr', 'i2'),
    ('mem_bank', 'i2'),
    ('mod', 'a16'),
    ('syn_th', 'f4'),
    ('dead_time_comp', 'i2'),
    ('polarity_l', 'i2'),
    ('polarity_f', 'i2'),
    ('polarity_p', 'i2'),
    ('linediv', 'i2'),
    ('accumulate', 'i2'),
    ('flbck_y', 'i4'),
    ('flbck_x', 'i4'),
    ('bord_u', 'i4'),
    ('bord_l', 'i4'),
    ('pix_time', 'f4'),
    ('pix_clk', 'i2'),
    ('trigger', 'i2'),
    ('scan_x', 'i4'),
    ('scan_y', 'i4'),
    ('scan_rx', 'i4'),
    ('scan_ry', 'i4'),
    ('fifo_typ', 'i2'),
    ('epx_div', 'i4'),
    ('mod_code', 'u2'),
    ('mod_fpga_ver', 'u2'),
    ('overflow_corr_factor', 'f4'),
    ('adc_zoom', 'i4'),
    ('cycles', 'i4'),
    ('StopInfo', MEASURE_STOP_INFO),
    ('FCSInfo', MEASURE_FCS_INFO),
    ('image_x', 'i4'),
    ('image_y', 'i4'),
    ('image_rx', 'i4'),
    ('image_ry', 'i4'),
    ('xy_gain', 'i2'),
    ('master_clock', 'i2'),
    ('adc_de', 'i2'),
    ('det', 'i2'),
    ('x_axis', 'i2'),
    ('MeasHISTInfo', HIST_INFO)]

BLOCK_HEADER = [
    ('block_no', 'i2'),
    ('data_offs', 'i4'),
    ('next_block_offs', 'i4'),
    ('block_type', 'u2'),
    ('meas_desc_block_no', 'i2'),
    ('lblock_no', 'u4'),
    ('block_length', 'u4')]

BLOCK_CREATION = {  # mode of creation
    0: 'NOT_USED',
    1: 'MEAS_DATA',
    2: 'FLOW_DATA',
    3: 'MEAS_DATA_FROM_FILE',
    4: 'CALC_DATA',
    5: 'SIM_DATA',
    8: 'FIFO_DATA',
    9: 'FIFO_DATA_FROM_FILE'}

BLOCK_CONTENT = {
    0x0: 'DECAY_BLOCK',
    0x10: 'PAGE_BLOCK',
    0x20: 'FCS_BLOCK',
    0x30: 'FIDA_BLOCK',
    0x40: 'FILDA_BLOCK',
    0x50: 'MCS_BLOCK',
    0x60: 'IMG_BLOCK'}

BLOCK_DTYPE = {  # data type
    0x000: numpy.dtype('<u2'),
    0x100: numpy.dtype('<u4'),
    0x200: numpy.dtype('<f4')}

HEADER_VALID = {
    0x1111: False,
    0x5555: True}

INFO_IDS = {
    b"SPC Setup Script File": "Setup script mode: setup only",
    b"SPC Setup & Data File": "Normal mode: setup + data",
    b"SPC DLL Data File": "DLL created: no setup, only data",
    b"SPC Flow Data File": "Continuous Flow mode: no setup, only data",
    b"SPC FCS Data File":
    "FIFO mode: setup, data blocks = Decay, FCS, FIDA, FILDA & MCS "
    "curves for each used routing channel"}

if sys.version_info[0] > 2:
    basestring = str
    bytes2str = lambda x: str(x, 'ascii')
else:
    bytes2str = str

if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/python3
import numpy
from multiprocessing.pool import Pool
import pickle
from inspect import getsource
from datetime import datetime
from lmfit import Parameters, minimize
from fepydas.libs.fit_functions import BiExponential, BiExponentialTail, Exponential, Gaussian, Lorentzian, Linear, TriExponential, Quadratic
from fepydas.datatypes.Data import Transformation
from scipy.signal import convolve

class Fit:
  def __init__(self, function):
    self.function = function

  def saveBinary(self, filename):
    f = open(filename,"bw")
    pickle.dump(self,f)
    f.close()

  def toTransformation(self):
    return Transformation(self.function,self.result.params.valuesdict())

  def initializeParameters(self, parameters: Parameters):
    self.parameters = parameters
 
  def residual(self, params, x, data=None, eps=None):
    parvals = params.valuesdict()
    if eps  is not None:
      return (data-self.function(x,**parvals))*eps
    else:
      return data-self.function(x,**parvals)

  def residualLog(self, params, x, data=None, eps=None):
    parvals = params.valuesdict()
    vals = self.function(x,**parvals)
    vals = numpy.maximum(vals, numpy.full_like(vals, 1E-9))
    if eps  is not None:
      return (numpy.log(data)-numpy.log(vals))*eps
    else:
      return (numpy.log(data)-numpy.log(vals))

  def convolve(self, signal, ref):
    return convolve(signal, ref, mode="full")[:len(signal)]

  def convolutedResidual(self, params, x, data=None, irf=None):
    parvals = params.valuesdict()
    return data - self.convolve(self.function(x,**parvals),irf)

  def fit(self, x, y, eps=None, nan_policy="raise", log=False):
    if log:
      self.result = minimize(self.residualLog, self.parameters, args=(x,y,eps), method="leastsq", nan_policy=nan_policy)
    else:
      self.result = minimize(self.residual, self.parameters, args=(x,y,eps), method="leastsq", nan_policy=nan_policy)
    return self.result

  def fitSpectrum(self, spectrum):
    return self.fit(spectrum.axis.values, spectrum.data.values)

  def batchFit(self, x, y):
    print("BatchFit",x.shape, y.shape)
    args = {}
    for i in range(y.shape[0]):
      args[i] = [x, y[i]]
    self.executeBatchFit(self.fit, args)

  def convolutedFit(self, x, y, irf):
    self.result = minimize(self.convolutedResidual, self.parameters, args =(x,y,irf))
    return self.result

  def convolutedBatchFit(self, x, y, irf):
    args = {}
    for i in range(y.shape[0]):
      args[i] = [x, y[i], irf]
    self.executeBatchFit(self.convolutedFit, args)

  def executeBatchFit(self, func, args):
    pool = Pool()
    jobs = {}
    for key in args.keys():
      jobs[key] = pool.apply_async(func, args[key])
    results = {}
    for key in args.keys():
      results[key]=jobs[key].get(1000) 
    self.results = results

  def batchEvaluate(self, x):
    data = numpy.ndarray(shape=(len(self.results.keys()), len(x)))
    for key in self.results.keys():
      pars = self.results[key].params.valuesdict()
      data[key,:] = self.function(x, **pars)
    return data

  def batchEvaluateConvolution(self, x, irf):
    data = self.batchEvaluate(x)
    for i in range(data.shape[0]):
      data[i,:] = self.convolve(data[i,:],irf)
    return data

  def evaluate(self, x):
    parvals = self.result.params.valuesdict()
    return self.function(x,**parvals)

  def evaluateInput(self, x):
    parvals = self.parameters.valuesdict()
    return self.function(x,**parvals)
  
  def evaluateConvolution(self, x, irf):
    return self.convolve(self.evaluate(x),irf)
  
  def startReport(self, filename):
    f = open(filename,"w") 
    f.write("Generated by Fit: {0}\n".format(self.__class__.__name__))
    f.write("Time: {0}\n".format(datetime.now()))
    f.write("Model: {0}\n".format(self.function.__name__))
    f.write("{0}\n".format("\n".join(getsource(self.function).split("\n")[1:-1])))
    f.write("Input Parameters: \n")
    for p in self.parameters.keys():
      f.write("  {0}:\t{1}\t[{2}:{3}]\n" .format(p,self.parameters[p].value,self.parameters[p].min,self.parameters[p].max))
    f.write("---\n")
    return f

  def saveReport(self, filename):
   f = self.startReport(filename)
   params = self.parameters.keys()
   f.write("Parameter\tValue\tError\n")
   for p in params:
     f.write("{0}\t{1}\t{2}\n".format(p,self.result.params[p].value,self.result.params[p].stderr))
   f.write("\nnfev\t{0}\tchisqr\t{1}\tredchi\t{2}\n".format(self.result.nfev,self.result.chisqr,self.result.redchi))
   f.close()

  def batchSaveReport(self, filename, labels):
    f = self.startReport(filename)
    keys = self.results.keys()
    params = self.parameters.keys()
    f.write("Dataset")
    for p in params:
      f.write("\t{0}\t{0}Error".format(p))
    f.write("\tnfev\tchisqr\tredchi")
    f.write("\n")
    for k in keys:
      f.write("{0}".format(labels[k]))
      for p in params:
        f.write("\t{0}\t{1}".format(self.results[k].params[p].value,self.results[k].params[p].stderr))
      f.write("\t{0}\t{1}\t{2}".format(self.results[k].nfev,self.results[k].chisqr,self.results[k].redchi))
      f.write("\n")
    f.close()

  def initializeAutoFromSpectrum(self, spectrum):
    self.initializeAuto(spectrum.axis.values, spectrum.data.values)

class LorentzianFit(Fit):
  def __init__(self):
    super().__init__(Lorentzian)

class SpectralLine(LorentzianFit):
  def __init__(self):
    super().__init__()
  def initializeAuto(self, x, y): 
    bg = (y[0]+y[-1])/2
    I = numpy.max(y)-bg
    x0 = x[numpy.argmax(y)]
    idx = numpy.where(y>bg+I/2)[0]
    fwhm = numpy.abs(x[idx[-1]] - x[idx[0]])
    params = Parameters()
    params.add("bg", value=bg)
    params.add("I", value=I)
    params.add("x0", value=x0)
    params.add("fwhm", value=fwhm)
    self.initializeParameters(params)

class GaussianFit(Fit):
  def __init__(self):
    super().__init__(Gaussian)

  def initializeAuto(self, x, y): 
    bg = (y[0]+y[-1])/2
    I = numpy.max(y)-bg
    x0 = x[numpy.argmax(y)]
    idx = numpy.where(y>bg+I/2)[0]
    fwhm = numpy.abs(x[idx[-1]] - x[idx[0]])
    params = Parameters()
    params.add("bg", value=bg)
    params.add("I", value=I)
    params.add("x0", value=x0)
    params.add("fwhm", value=fwhm+x0/1000)
    self.initializeParameters(params)

class LimitedGaussianFit(GaussianFit):
  def __init__(self):
    super().__init__()
  
  def initializeAutoLimited(self, x, y, center, range, thresh=1):
    if (len(numpy.where(y == numpy.max(y))[0])) > 1:
      return False, False
    lowerIdx = (numpy.abs(x - (center-range))).argmin()
    higherIdx = (numpy.abs(x - (center+range))).argmin()
    if numpy.abs(higherIdx-lowerIdx) < 15:
      return False, False
    if higherIdx < lowerIdx:
      t = lowerIdx
      lowerIdx = higherIdx
      higherIdx = t
    x = x[lowerIdx:higherIdx]
    y = y[lowerIdx:higherIdx]
    if numpy.amax(y) <= thresh:
      return False, False
    self.initializeAuto(x, y)
    return x, y  

class LinearFit(Fit):
  def __init__(self):
    super().__init__(Linear)

class CalibrationFit(LinearFit):
  def __init__(self):
    super().__init__()
  def initializeAuto(self):
    params = Parameters()
    params.add("a",value=1)
    params.add("b",value=0)
    self.initializeParameters(params)

class QuadraticFit(Fit):
  def __init__(self):
    super().__init__(Quadratic)

class QuadraticCalibrationFit(QuadraticFit):
  def __init__(self):
    super().__init__()
  def initializeAuto(self):
    params = Parameters()
    params.add("a", value = 0)
    params.add("b", value = 1)
    params.add("c", value = 0)
    self.initializeParameters(params)  

class AutomaticCalibration(QuadraticCalibrationFit):
  def __init__(self, spectrum, references, threshold=10, width=10):
    super().__init__()
    peaks = spectrum.identifyPeaks(threshold=threshold, width=width)
    SpectralFit = SpectralLine()
    peakVals = []
    peakErrs = []
    for peak in peaks:
      x,y = spectrum.axis.values[peak[0]:peak[1]], spectrum.data.values[peak[0]:peak[1]]
      SpectralFit.initializeAuto(x,y)
      SpectralFit.fit(x,y)
      peakVals.append(SpectralFit.result.params["x0"].value)
      peakErrs.append(SpectralFit.result.params["x0"].stderr)
    references = numpy.array(references,dtype=numpy.float64)
    peakVals = numpy.array(peakVals,dtype=numpy.float64)
    peakErrs = numpy.array(peakErrs,dtype=numpy.float64)
    idx = numpy.where(~numpy.isnan(references))[0]
    print("Calibration with ",references,peakVals,idx)
    self.initializeAuto()
    self.fit(peakVals[idx], references[idx])
  

class ExponentialFit(Fit):
  def __init__(self):
    super().__init__(Exponential)
  def initialize(self,bg,I,x0,tau,rise):
    params = Parameters()
    params.add("bg",value=bg)
    params.add("I",value=I,min=0)
    params.add("x0",value=x0)
    params.add("tau",value=tau,min=0)
    params.add("rise",value=rise,min=0)
    self.initializeParameters(params)

  def initializeAuto(self,x,y):
    params = Parameters()
    params.add("bg",value=y[0])
    params.add("I",value=numpy.max(y)-y[0],min=0)
    #params.add("I_rise",value=numpy.max(y)-y[0],min=0)
    params.add("x0",value=x[numpy.argmax(y)])
    params.add("tau",value=1,min=0)
    params.add("rise",value=0.001,min=0)
    self.initializeParameters(params)

class BiExponentialFit(Fit):
  def __init__(self):
    super().__init__(BiExponential)
  def initializeAuto(self,x,y):
    params = Parameters()
    params.add("bg",value=y[0])
    params.add("I_1",value=(numpy.max(y)-y[0])/2,min=0)
    params.add("I_2",value=(numpy.max(y)-y[0])/2,min=0)
    params.add("x0",value=x[numpy.argmax(y)])
    params.add("tau_1",value=5,min=0)
    params.add("tau_2",value=50,min=0)
    params.add("rise",value=0.001,min=0)
    self.initializeParameters(params)

  def initialize(self,I_1, I_2, tau_1, tau_2, x0=0, rise= 0.001):
    params = Parameters()
    params.add("bg",value=0)
    params.add("I_1",value=I_1,min=0)
    params.add("I_2",value=I_2,min=0)
    params.add("x0",value=x0)
    params.add("tau_1",value=tau_1,min=0)
    params.add("tau_2",value=tau_2,min=0)
    params.add("rise",value=rise,min=0)
    self.initializeParameters(params)

class TriExponentialFit(Fit):
  def __init__(self):
    super().__init__(TriExponential)
  def initializeAuto(self,x,y):
    params = Parameters()
    params.add("bg",value=y[0])
    params.add("I_1",value=(numpy.max(y)-y[0])/3,min=0)
    params.add("I_2",value=(numpy.max(y)-y[0])/3,min=0)
    params.add("I_3",value=(numpy.max(y)-y[0])/3,min=0)
    params.add("x0",value=x[numpy.argmax(y)])
    params.add("tau_1",value=5,min=0)
    params.add("tau_2",value=50,min=0)
    params.add("tau_3",value=500,min=0)
    params.add("rise",value=0.001,min=0)
    self.initializeParameters(params)

  def initialize(self,I_1, I_2, I_3, tau_1, tau_2, tau_3,x0=0, rise= 0.001):
    params = Parameters()
    params.add("bg",value=0)
    params.add("I_1",value=I_1,min=0)
    params.add("I_2",value=I_2,min=0)
    params.add("I_3",value=I_3,min=0)
    params.add("x0",value=x0)
    params.add("tau_1",value=tau_1,min=0)
    params.add("tau_2",value=tau_2,min=0)
    params.add("tau_3",value=tau_3,min=0)
    params.add("rise",value=rise,min=0)
    self.initializeParameters(params)


    
class BiExponentialTailFit(Fit):
  def __init__(self):
    super().__init__(BiExponentialTail)
  def initializeAuto(self,x,y):
    params = Parameters()
    params.add("bg",value=y[-1])
    params.add("I_1",value=(numpy.max(y)-y[-1])/2,min=0)
    params.add("I_2",value=(numpy.max(y)-y[-1])/2,min=0)
    params.add("tau_1",value=5,min=0)
    params.add("tau_2",value=50,min=0)
    self.initializeParameters(params)
  def initialize(self, I_1, I_2, tau_1, tau_2):
    params = Parameters()
    params.add("bg",value=0)
    params.add("I_1",value=I_1,min=0)
    params.add("I_2",value=I_2,min=0)
    params.add("tau_1",value=tau_1,min=0)
    params.add("tau_2",value=tau_2,min=0)
    self.initializeParameters(params)



#!/usr/bin/python3
from setuptools import setup, find_packages
import pathlib
# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README").read_text()
setup(
  name = "fepydas",
  version = "0.0.25",
  author = "Felix Nippert",
  author_email = "felix@physik.tu-berlin.de",
  description="felix' python data analysis suite",
  long_description=README,
  long_description_content_type="text/markdown",
  url="https://git.tu-berlin.de/nippert/fepydas",
  packages = find_packages(),
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Programming Language :: Python :: 3',
  ],
  install_requires=['numpy','scipy','lmfit','matplotlib','scikit-learn', 'pyamg'],
  python_requires='>=3',
)



#!/usr/bin/python3
import numpy
from fepydas.libs.sdtfile import SdtFile

def PicoHarpHistogramIntegrationTime(filename):
    file = open(filename, 'br')
    bytes = file.read(1000000000)
    from struct import unpack,  calcsize
    skip = 0
    fmt_txtHdr="8s8s"
    size_txtHdr = calcsize(fmt_txtHdr)
    data_txtHdr = unpack(fmt_txtHdr,  bytes[skip:skip+size_txtHdr])
    skip+=size_txtHdr
    fmt_Hdr="32s3i1I"
    size_Hdr = calcsize(fmt_Hdr)
    NumberOfCurves = 0
    HistResDscr = {}
    while True:
        data_Hdr = unpack(fmt_Hdr,  bytes[skip:skip+size_Hdr])
        skip+=size_Hdr
        if data_Hdr[2] & 65535 == 65535:
            skip+=data_Hdr[3]
        if data_Hdr[0] == b'HistResDscr_MDescStopAfter\x00\x00\x00\x00\x00\x00':
        #if data_Hdr[0] == b'MeasDesc_AcquisitionTime\x00\x00\x00\x00\x00\x00\x00\x00':
            time = data_Hdr[3]
            break
        if data_Hdr[0] == b'Header_End\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00':
            break
    file.close()
    return time  

def PicoHarpHistogram(filename): 
    file = open(filename, 'br')
    bytes = file.read(1000000000)
    from struct import unpack,  calcsize
    skip = 0
    fmt_txtHdr="8s8s"
    size_txtHdr = calcsize(fmt_txtHdr)
    data_txtHdr = unpack(fmt_txtHdr,  bytes[skip:skip+size_txtHdr])
    skip+=size_txtHdr
    fmt_Hdr="32s3i1I"
    size_Hdr = calcsize(fmt_Hdr)
    NumberOfCurves = 0
    HistResDscr = {}
    while True:
        data_Hdr = unpack(fmt_Hdr,  bytes[skip:skip+size_Hdr])
        skip+=size_Hdr
        if data_Hdr[2] & 65535 == 65535:
            skip+=data_Hdr[3]
        if data_Hdr[0] == b'Header_End\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00':
            break   
        if data_Hdr[0] == b'HistoResult_NumberOfCurves\x00\x00\x00\x00\x00\x00':
            NumberOfCurves = data_Hdr[3]
        if data_Hdr[0] == b'HistResDscr_DataOffset\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00':
            curveId = data_Hdr[1]
            value = data_Hdr[3]
            if not curveId in HistResDscr.keys():
                HistResDscr[curveId] = {}
            HistResDscr[curveId] ["DataOffset"] = value
        if data_Hdr[0] == b'HistResDscr_HistogramBins\x00\x00\x00\x00\x00\x00\x00':
            curveId = data_Hdr[1]
            value = data_Hdr[3]
            if not curveId in HistResDscr.keys():
                HistResDscr[curveId] = {}
            HistResDscr[curveId] ["HistogramBins"] = value
        if data_Hdr[0] == b'HistResDscr_MDescResolution\x00\x00\x00\x00\x00':
            curveId = data_Hdr[1]
            fmt_Float="1d"
            size_Float = calcsize(fmt_Float)
            data_Float = data_Hdr = unpack(fmt_Float,  bytes[skip-size_Float:skip])
            value = data_Float[0]
            if not curveId in HistResDscr.keys():
                HistResDscr[curveId] = {}
            HistResDscr[curveId]["MDescResolution"] = value
    data = numpy.zeros(shape=(HistResDscr[0]["HistogramBins"]),dtype=numpy.int64)
    print("Number of Curves: {0}".format(NumberOfCurves))
    for i in range(NumberOfCurves):
        offset = HistResDscr[i]["DataOffset"]
        fmt_Histogram = "{0}I".format(HistResDscr[i]["HistogramBins"])
        size_Histogram = calcsize(fmt_Histogram)
        Histogram = unpack(fmt_Histogram, bytes[offset:offset+size_Histogram])
        data += Histogram
    print("Bins: {0} Resolution: {1}".format(HistResDscr[0]["HistogramBins"],HistResDscr[0]["MDescResolution"]))
    time = numpy.arange(0,  (HistResDscr[0]["HistogramBins"])*HistResDscr[0]["MDescResolution"] /1e-9,  HistResDscr[0]["MDescResolution"]/1e-9)
    file.close()
    return time, data

def TimeHarpHistogram(filename):
    file = open(filename, 'br')
    bytes = file.read(100000000)
    from struct import unpack,  calcsize
    skip = 0
    fmt_txtHdr="16s6s18s12s18s2s256s"
    size_txtHdr = calcsize(fmt_txtHdr)
    skip+=size_txtHdr
    
    fmt_binHdr="6id2iI3i4I16i9f4i20s"
    size_binHdr = calcsize(fmt_binHdr)
    skip+=size_binHdr
    
    fmt_mainHardwareHdr = "16s8s16s2id12i"
    size_mainHardwareHdr = calcsize(fmt_mainHardwareHdr)
    data_mainHardwareHdr = unpack(fmt_mainHardwareHdr,  bytes[skip:skip+size_mainHardwareHdr])
    inpChansPresent = data_mainHardwareHdr[7]
    skip+=size_mainHardwareHdr
    
    fmt_inputChannelSettings= "4i"
    size_inputChannelsettings=inpChansPresent*calcsize(fmt_inputChannelSettings)
    skip+=size_inputChannelsettings
    
    fmt_curveHdr = "iI16s8s16s2id20id4i3f3iQ2i"
    size_curveHdr = calcsize(fmt_curveHdr)
    data_curveHdr = unpack(fmt_curveHdr,  bytes[skip:skip+size_curveHdr])
    bins = data_curveHdr[40]
    offset = data_curveHdr[41]
    resolution = data_curveHdr[28]
    print("Acquisition Time: {0}s".format(data_curveHdr[31]/1000))
    print("Resolution: {0}ns".format(resolution/1000))
    
    fmt_counts = "{0}I".format(bins)
    size_counts = calcsize(fmt_counts)
    data_counts = unpack(fmt_counts,  bytes[offset:offset+size_counts])
    data = numpy.array(data_counts)
    time = numpy.arange(0,  (bins)*resolution/1000,  resolution/1000)

    return time, data

def BeckerHicklHistogram(filename):
    sdt = SdtFile(filename)
    x = numpy.ndarray(shape=(len(sdt.times[0][:])))
    y = numpy.ndarray(shape=(len(sdt.times[0][:])))
    print(len(sdt.times[0]))
    for i in range(len(sdt.times[0][:])):
        x[i] = numpy.abs(sdt.times[0][i]*1000000000)
        y[i] = sdt.data[0][0][i]
    return x, y
